import logging

import hypothesis.strategies as hst
import numpy as np
import pytest
from hypothesis import HealthCheck, given, settings
from numpy.testing import assert_allclose, assert_array_equal

from qcodes.data.location import FormatLocation
from qcodes.instrument.channel import ChannelList
from qcodes.instrument.parameter import Parameter
from qcodes.loops import Loop
from qcodes.tests.instrument_mocks import DummyChannel, DummyChannelInstrument
from qcodes.utils.validators import Numbers


@pytest.fixture(scope='function', name='dci')
def _make_dci():

    dci = DummyChannelInstrument(name='dci')
    try:
        yield dci
    finally:
        dci.close()


def test_channels_call_function(dci, caplog):
    """
    Test that dci.channels.some_function() calls
    some_function on each of the channels
    """
    with caplog.at_level(logging.DEBUG,
                         logger='qcodes.tests.instrument_mocks'):
        caplog.clear()
        dci.channels.log_my_name()
        mssgs = [rec.message for rec in caplog.records]
        names = [ch.name.replace('dci_', '') for ch in dci.channels]
        assert mssgs == names


def test_channels_get(dci):

    temperatures = dci.channels.temperature.get()
    assert len(temperatures) == 6


@settings(suppress_health_check=(HealthCheck.function_scoped_fixture,))
@given(value=hst.floats(0, 300), channel=hst.integers(0, 3))
def test_channel_access_is_identical(dci, value, channel):
    channel_to_label = {0: 'A', 1: 'B', 2: 'C', 3: "D"}
    label = channel_to_label[channel]
    channel_via_label = getattr(dci, label)
    channel_via_name = dci.channels.get_channel_by_name(f"Chan{label}")
    # set via labeled channel
    channel_via_label.temperature(value)
    assert channel_via_label.temperature() == value
    assert channel_via_name.temperature() == value
    assert dci.channels[channel].temperature() == value
    assert dci.channels.temperature()[channel] == value
    # reset via channel name
    channel_via_name.temperature(0)
    assert channel_via_label.temperature() == 0
    assert channel_via_name.temperature() == 0
    assert dci.channels[channel].temperature() == 0
    assert dci.channels.temperature()[channel] == 0
    # set via index into list
    dci.channels[channel].temperature(value)
    assert channel_via_label.temperature() == value
    assert channel_via_name.temperature() == value
    assert dci.channels[channel].temperature() == value
    assert dci.channels.temperature()[channel] == value
    # it's not possible to set via dci.channels.temperature
    # as this is a multi parameter that currently does not support set.


def test_add_channel(dci):
    n_channels = len(dci.channels)
    name = 'foo'
    channel = DummyChannel(dci, 'Chan'+name, name)
    dci.channels.append(channel)
    dci.add_submodule(name, channel)

    assert len(dci.channels) == n_channels+1

    dci.channels.lock()
    # after locking the channels it's not possible to add any more channels
    with pytest.raises(AttributeError):
        name = 'bar'
        channel = DummyChannel(dci, 'Chan' + name, name)
        dci.channels.append(channel)
    assert len(dci.channels) == n_channels + 1


def test_add_channels_from_generator(dci):
    n_channels = len(dci.channels)
    names = ('foo', 'bar', 'foobar')
    channels = (DummyChannel(dci, 'Chan'+name, name)
                for name in names)
    dci.channels.extend(channels)

    assert len(dci.channels) == n_channels + len(names)


def test_add_channels_from_tuple(dci):
    n_channels = len(dci.channels)
    names = ('foo', 'bar', 'foobar')
    channels = tuple(DummyChannel(dci, 'Chan'+name, name)
                     for name in names)
    dci.channels.extend(channels)

    assert len(dci.channels) == n_channels + len(names)


def test_extend_then_remove(dci):
    n_channels = len(dci.channels)
    names = ('foo', 'bar', 'foobar')
    channels = [DummyChannel(dci, 'Chan' + name, name)
                for name in names]
    dci.channels.extend(channels)

    assert len(dci.channels) == n_channels + len(names)
    last_channel = dci.channels[-1]
    dci.channels.remove(last_channel)
    assert last_channel not in dci.channels
    assert len(dci.channels) == n_channels + len(names) - 1


def test_insert_channel(dci):
    n_channels = len(dci.channels)
    name = 'foo'
    channel = DummyChannel(dci, 'Chan'+name, name)
    dci.channels.insert(1, channel)
    dci.add_submodule(name, channel)

    assert len(dci.channels) == n_channels+1
    assert dci.channels[1] is channel
    dci.channels.lock()
    # after locking the channels it's not possible to add any more channels
    with pytest.raises(AttributeError):
        name = 'bar'
        channel = DummyChannel(dci, 'Chan' + name, name)
        dci.channels.insert(2, channel)
    assert len(dci.channels) == n_channels + 1


def test_clear_channels(dci):
    channels = dci.channels
    channels.clear()
    assert len(channels) == 0


def test_clear_locked_channels(dci):
    channels = dci.channels
    original_length = len(channels)
    channels.lock()
    with pytest.raises(AttributeError):
        channels.clear()
    assert len(channels) == original_length


def test_remove_channel(dci):
    channels = dci.channels
    chan_a = dci.A
    original_length = len(channels.temperature())
    channels.remove(chan_a)
    with pytest.raises(AttributeError):
        getattr(channels, chan_a.short_name)
    assert len(channels) == original_length-1
    assert len(channels.temperature()) == original_length-1


def test_remove_locked_channel(dci):
    channels = dci.channels
    chan_a = dci.A
    channels.lock()
    with pytest.raises(AttributeError):
        channels.remove(chan_a)


def test_remove_tupled_channel(dci):
    channel_tuple = tuple(
        DummyChannel(dci, f'Chan{C}', C)
        for C in ('A', 'B', 'C', 'D', 'E', 'F')
    )
    channels = ChannelList(dci,
                           "TempSensorsTuple",
                           DummyChannel,
                           channel_tuple,
                           snapshotable=False)
    chan_a = channels.ChanA
    with pytest.raises(AttributeError):
        channels.remove(chan_a)


@settings(suppress_health_check=(HealthCheck.function_scoped_fixture,))
@given(setpoints=hst.lists(hst.floats(0, 300), min_size=4, max_size=4))
def test_combine_channels(dci, setpoints):
    assert len(dci.channels) == 6

    mychannels = dci.channels[0:2] + dci.channels[4:]

    assert len(mychannels) == 4
    assert mychannels[0] is dci.A
    assert mychannels[1] is dci.B
    assert mychannels[2] is dci.E
    assert mychannels[3] is dci.F

    for i, chan in enumerate(mychannels):
        chan.temperature(setpoints[i])

    expected = tuple(setpoints[0:2] + [0, 0] + setpoints[2:])
    assert dci.channels.temperature() == expected


@settings(suppress_health_check=(HealthCheck.function_scoped_fixture,))
@given(start=hst.integers(-8, 7), stop=hst.integers(-8, 7),
       step=hst.integers(1, 7))
def test_access_channels_by_slice(dci, start, stop, step):
    names = ('A', 'B', 'C', 'D', 'E', 'F', 'G', 'H')
    channels = tuple(DummyChannel(dci,
                                  'Chan'+name, name) for name in names)
    chlist = ChannelList(dci, 'channels',
                         DummyChannel, channels)
    if stop < start:
        step = -step
    myslice = slice(start, stop, step)
    mychans = chlist[myslice]
    expected_channels = names[myslice]
    for chan, exp_chan in zip(mychans, expected_channels):
        assert chan.name == f'dci_Chan{exp_chan}'


@settings(suppress_health_check=(HealthCheck.function_scoped_fixture,))
@given(myindexs=hst.lists(elements=hst.integers(-8, 7), min_size=1))
def test_access_channels_by_tuple(dci, myindexs):
    names = ('A', 'B', 'C', 'D', 'E', 'F', 'G', 'H')
    mytuple = tuple(myindexs)
    channels = tuple(DummyChannel(dci,
                                  'Chan'+name, name) for name in names)
    chlist = ChannelList(dci, 'channels',
                         DummyChannel, channels)

    mychans = chlist[mytuple]
    for chan, chanindex in zip(mychans, mytuple):
        assert chan.name == f"dci_Chan{names[chanindex]}"


@settings(suppress_health_check=(HealthCheck.function_scoped_fixture,))
@given(myindexs=hst.lists(elements=hst.integers(0, 7), min_size=2))
def test_access_channels_by_name(dci, myindexs):
    names = ("A", "B", "C", "D", "E", "F", "G", "H")
    channels = tuple(DummyChannel(dci, "Chan" + name, name) for name in names)
    chlist = ChannelList(dci, "channels", DummyChannel, channels)

    channel_names = (f"Chan{names[i]}" for i in myindexs)
    mychans = chlist.get_channel_by_name(*channel_names)
    for chan, chanindex in zip(mychans, myindexs):
        assert chan.name == f'dci_Chan{names[chanindex]}'


def test_names(dci):
    ex_inst_name = 'dci'
    for channel in dci.channels:
        sub_channel = DummyChannel(channel, 'subchannel', 'subchannel')
        channel.add_submodule('somesubchannel', sub_channel)
    assert dci.name == ex_inst_name
    assert dci.full_name == ex_inst_name
    assert dci.short_name == ex_inst_name
    assert dci.name_parts == [ex_inst_name]

    # Parameters directly on instrument
    assert dci.IDN.name == 'IDN'
    assert dci.IDN.full_name == f"{ex_inst_name}_IDN"
    for chan, name in zip(dci.channels,
                          ['A', 'B', 'C', 'D', 'E', 'F']):
        ex_chan_name = f"Chan{name}"
        ex_chan_full_name = f"{ex_inst_name}_{ex_chan_name}"

        assert chan.short_name == ex_chan_name
        assert chan.name == ex_chan_full_name
        assert chan.full_name == ex_chan_full_name
        assert chan.name_parts == [ex_inst_name, ex_chan_name]

        ex_param_name = 'temperature'
        assert chan.temperature.name == ex_param_name
        assert chan.temperature.full_name ==\
               f'{ex_chan_full_name}_{ex_param_name}'
        assert chan.temperature.short_name == ex_param_name
        assert chan.temperature.name_parts == [ex_inst_name, ex_chan_name,
                                               ex_param_name]

        ex_subchan_name = f"subchannel"
        ex_subchan_full_name = f"{ex_chan_full_name}_{ex_subchan_name}"

        assert chan.somesubchannel.short_name == ex_subchan_name
        assert chan.somesubchannel.name == ex_subchan_full_name
        assert chan.somesubchannel.full_name == ex_subchan_full_name
        assert chan.somesubchannel.name_parts == [ex_inst_name,
                                                  ex_chan_name,
                                                  ex_subchan_name]

        assert chan.somesubchannel.temperature.name == ex_param_name
        assert chan.somesubchannel.temperature.full_name ==\
               f'{ex_subchan_full_name}_{ex_param_name}'
        assert chan.somesubchannel.temperature.short_name == ex_param_name
        assert chan.somesubchannel.temperature.name_parts == \
               [ex_inst_name, ex_chan_name, ex_subchan_name, ex_param_name]


def test_loop_simple(dci):
    loc_fmt = 'data/{date}/#{counter}_{name}_{date}_{time}'
    rcd = {'name': 'loopSimple'}
    loc_provider = FormatLocation(fmt=loc_fmt, record=rcd)
    loop = Loop(dci.channels[0].temperature.sweep(0, 300, 10),
                0.001).each(dci.A.temperature)
    data = loop.run(location=loc_provider)
    assert_array_equal(data.dci_ChanA_temperature_set.ndarray,
                       data.dci_ChanA_temperature.ndarray)


def test_loop_measure_all_channels(dci):
    p1 = Parameter(name='p1', vals=Numbers(-10, 10), get_cmd=None,
                   set_cmd=None)
    loc_fmt = 'data/{date}/#{counter}_{name}_{date}_{time}'
    rcd = {'name': 'allChannels'}
    loc_provider = FormatLocation(fmt=loc_fmt, record=rcd)
    loop = Loop(p1.sweep(-10, 10, 1), 1e-6).\
        each(dci.channels.temperature)
    data = loop.run(location=loc_provider)
    assert data.p1_set.ndarray.shape == (21, )
    assert len(data.arrays) == 7
    for chan in ['A', 'B', 'C', 'D', 'E', 'F']:
        assert getattr(
            data,
            f'dci_Chan{chan}_temperature'
        ).ndarray.shape == (21,)


def test_loop_measure_channels_individually(dci):
    p1 = Parameter(name='p1', vals=Numbers(-10, 10), get_cmd=None,
                   set_cmd=None)
    loc_fmt = 'data/{date}/#{counter}_{name}_{date}_{time}'
    rcd = {'name': 'channelsIndividually'}
    loc_provider = FormatLocation(fmt=loc_fmt, record=rcd)
    loop = Loop(p1.sweep(-10, 10, 1), 1e-6).each(dci.
                                                 channels[0].temperature,
                                                 dci.
                                                 channels[1].temperature,
                                                 dci.
                                                 channels[2].temperature,
                                                 dci.
                                                 channels[3].temperature)
    data = loop.run(location=loc_provider)
    assert data.p1_set.ndarray.shape == (21, )
    for chan in ['A', 'B', 'C', 'D']:
        assert getattr(
            data, f'dci_Chan{chan}_temperature'
        ).ndarray.shape == (21,)


@given(values=hst.lists(hst.floats(0, 300), min_size=4, max_size=4))
@settings(max_examples=10, deadline=None, suppress_health_check=(HealthCheck.function_scoped_fixture,))
def test_loop_measure_channels_by_name(dci, values):
    p1 = Parameter(name='p1', vals=Numbers(-10, 10), get_cmd=None,
                   set_cmd=None)
    for i in range(4):
        dci.channels[i].temperature(values[i])
    loc_fmt = 'data/{date}/#{counter}_{name}_{date}_{time}'
    rcd = {'name': 'channelsByName'}
    loc_provider = FormatLocation(fmt=loc_fmt, record=rcd)
    loop = Loop(p1.sweep(-10, 10, 1), 1e-6).each(
        dci.A.temperature,
        dci.B.temperature,
        dci.C.temperature,
        dci.D.temperature
    )
    data = loop.run(location=loc_provider)
    assert data.p1_set.ndarray.shape == (21, )
    for i, chan in enumerate(['A', 'B', 'C', 'D']):
        assert getattr(
            data, f'dci_Chan{chan}_temperature'
        ).ndarray.shape == (21,)
        assert getattr(
            data, f'dci_Chan{chan}_temperature'
        ).ndarray.max() == values[i]
        assert getattr(
            data, f'dci_Chan{chan}_temperature'
        ).ndarray.min() == values[i]


@given(loop_channels=hst.lists(hst.integers(0, 3), min_size=2, max_size=2,
                               unique=True),
       measure_channel=hst.integers(0, 3))
@settings(max_examples=10, deadline=800,
          suppress_health_check=(HealthCheck.function_scoped_fixture,))
def test_nested_loop_over_channels(dci, loop_channels, measure_channel):
    channel_to_label = {0: 'A', 1: 'B', 2: 'C', 3: "D"}
    loc_fmt = 'data/{date}/#{counter}_{name}_{date}_{time}'
    rcd = {'name': 'nestedLoopOverChannels'}
    loc_provider = FormatLocation(fmt=loc_fmt, record=rcd)
    loop = Loop(dci.channels[loop_channels[0]].temperature.
                sweep(0, 10, 0.5))
    loop = loop.loop(dci.channels[loop_channels[1]].temperature.
                     sweep(50, 51, 0.1))
    loop = loop.each(dci.channels[measure_channel].temperature)
    data = loop.run(location=loc_provider)

    assert getattr(
        data,
        f'dci_Chan{channel_to_label[loop_channels[0]]}_temperature_set'
    ).ndarray.shape == (21,)
    assert getattr(
        data,
        f'dci_Chan{channel_to_label[loop_channels[1]]}_temperature_set'
    ).ndarray.shape == (21, 11,)
    assert getattr(
        data,
        f'dci_Chan{channel_to_label[measure_channel]}_temperature'
    ).ndarray.shape == (21, 11)

    assert_array_equal(getattr(
        data,
        f'dci_Chan{channel_to_label[loop_channels[0]]}_temperature_set'
    ).ndarray, np.arange(0, 10.1, 0.5))

    expected_array = np.repeat(np.arange(50, 51.01, 0.1).reshape(1, 11),
                               21, axis=0)
    array = getattr(
        data,
        f'dci_Chan{channel_to_label[loop_channels[1]]}_temperature_set'
    ).ndarray
    assert_allclose(array, expected_array)


def test_loop_slicing_multiparameter_raises(dci):
    with pytest.raises(NotImplementedError):
        loop = Loop(dci.A.temperature.sweep(0, 10, 1), 0.1)
        loop.each(dci.channels[0:2].dummy_multi_parameter).run()


def test_loop_multiparameter_by_name(dci):
    loc_fmt = 'data/{date}/#{counter}_{name}_{date}_{time}'
    rcd = {'name': 'multiParamByName'}
    loc_provider = FormatLocation(fmt=loc_fmt, record=rcd)
    loop = Loop(dci.A.temperature.sweep(0, 10, 1), 0.1)
    data = loop.each(dci.A.dummy_multi_parameter)\
        .run(location=loc_provider)
    _verify_multiparam_data(data)
    assert 'multi_setpoint_param_this_setpoint_set' in data.arrays.keys()


def test_loop_multiparameter_by_index(dci):
    loc_fmt = 'data/{date}/#{counter}_{name}_{date}_{time}'
    rcd = {'name': 'loopByIndex'}
    loc_provider = FormatLocation(fmt=loc_fmt, record=rcd)
    loop = Loop(dci.channels[0].temperature.sweep(0, 10, 1),
                0.1)
    data = loop.each(dci.A.dummy_multi_parameter)\
        .run(location=loc_provider)
    _verify_multiparam_data(data)


def test_loop_slicing_arrayparameter(dci):
    loc_fmt = 'data/{date}/#{counter}_{name}_{date}_{time}'
    rcd = {'name': 'loopSlicing'}
    loc_provider = FormatLocation(fmt=loc_fmt, record=rcd)
    loop = Loop(dci.A.temperature.sweep(0, 10, 1), 0.1)
    data = loop.each(dci.channels[0:2].dummy_array_parameter)\
        .run(location=loc_provider)
    _verify_array_data(data, channels=('A', 'B'))


def test_loop_arrayparameter_by_name(dci):
    loc_fmt = 'data/{date}/#{counter}_{name}_{date}_{time}'
    rcd = {'name': 'arrayParamByName'}
    loc_provider = FormatLocation(fmt=loc_fmt, record=rcd)
    loop = Loop(dci.A.temperature.sweep(0, 10, 1), 0.1)
    data = loop.each(dci.A.dummy_array_parameter)\
        .run(location=loc_provider)
    _verify_array_data(data)


def test_loop_arrayparameter_by_index(dci):
    loc_fmt = 'data/{date}/#{counter}_{name}_{date}_{time}'
    rcd = {'name': 'arrayParamByIndex'}
    loc_provider = FormatLocation(fmt=loc_fmt, record=rcd)
    loop = Loop(dci.channels[0].temperature.sweep(0, 10, 1),
                0.1)
    data = loop.each(dci.A.dummy_array_parameter)\
        .run(location=loc_provider)
    _verify_array_data(data)


def test_root_instrument(dci):
    assert dci.root_instrument is dci
    for channel in dci.channels:
        assert channel.root_instrument is dci
        for parameter in channel.parameters.values():
            assert parameter.root_instrument is dci


def _verify_multiparam_data(data):
    assert 'multi_setpoint_param_this_setpoint_set' in data.arrays.keys()
    assert_array_equal(
        data.arrays['multi_setpoint_param_this_setpoint_set'].ndarray,
        np.repeat(np.arange(5., 10).reshape(1, 5), 11, axis=0)
    )
    assert 'dci_ChanA_multi_setpoint_param_this' in data.arrays.keys()
    assert_array_equal(
        data.arrays['dci_ChanA_multi_setpoint_param_this'].ndarray,
        np.zeros((11, 5))
    )
    assert 'dci_ChanA_multi_setpoint_param_this' in data.arrays.keys()
    assert_array_equal(
        data.arrays['dci_ChanA_multi_setpoint_param_that'].ndarray,
        np.ones((11, 5))
    )
    assert 'dci_ChanA_temperature_set' in data.arrays.keys()
    assert_array_equal(
        data.arrays['dci_ChanA_temperature_set'].ndarray,
        np.arange(0, 10.1, 1)
    )


def _verify_array_data(data, channels=('A',)):
    assert 'array_setpoint_param_this_setpoint_set' in data.arrays.keys()
    assert_array_equal(
        data.arrays['array_setpoint_param_this_setpoint_set'].ndarray,
        np.repeat(np.arange(5., 10).reshape(1, 5), 11, axis=0)
    )
    for channel in channels:
        aname = f'dci_Chan{channel}_dummy_array_parameter'
        assert aname in data.arrays.keys()
        assert_array_equal(data.arrays[aname].ndarray, np.ones((11, 5))+1)
    assert 'dci_ChanA_temperature_set' in data.arrays.keys()
    assert_array_equal(
        data.arrays['dci_ChanA_temperature_set'].ndarray,
        np.arange(0, 10.1, 1)
    )



import time
from math import floor

import pytest

from qcodes.dataset.data_set import (
    DataSet,
    load_by_counter,
    load_by_guid,
    load_by_id,
    load_by_run_spec,
    new_data_set,
)
from qcodes.dataset.data_set_info import get_run_attributes
from qcodes.dataset.descriptions.rundescriber import RunDescriber
from qcodes.dataset.experiment_container import new_experiment
from qcodes.dataset.sqlite.queries import (
    get_experiment_attributes_by_exp_id,
    get_guids_from_run_spec,
    get_raw_run_attributes,
    raw_time_to_str_time,
)


@pytest.mark.usefixtures("experiment")
def test_load_by_id():
    ds = new_data_set("test-dataset")
    run_id = ds.run_id
    ds.mark_started()
    ds.mark_completed()

    loaded_ds = load_by_id(run_id)
    assert ds.started is True
    assert ds.pristine is False
    assert ds.running is False
    assert loaded_ds.completed is True
    assert loaded_ds.exp_id == 1

    ds = new_data_set("test-dataset-unfinished")
    run_id = ds.run_id

    loaded_ds = load_by_id(run_id)
    assert ds.pristine is True
    assert ds.running is False
    assert ds.started is False
    assert loaded_ds.completed is False
    assert loaded_ds.exp_id == 1

    # let's take a run number that is not in the temporary test database file
    non_existing_run_id = run_id + 1
    with pytest.raises(ValueError, match=f"Run with run_id "
                                         f"{non_existing_run_id} does not "
                                         f"exist in the database"):
        _ = load_by_id(non_existing_run_id)


@pytest.mark.usefixtures("empty_temp_db")
def test_load_by_counter():
    exp = new_experiment(name="for_loading", sample_name="no_sample")
    ds = new_data_set("my_first_ds")

    loaded_ds = load_by_counter(exp.exp_id, 1)

    assert loaded_ds.pristine is True
    assert loaded_ds.running is False
    assert loaded_ds.started is False
    assert loaded_ds.completed is False

    ds.mark_started()
    ds.mark_completed()

    loaded_ds = load_by_counter(exp.exp_id, 1)

    assert loaded_ds.pristine is False
    assert loaded_ds.started is True
    assert loaded_ds.running is False
    assert loaded_ds.completed is True


@pytest.mark.usefixtures("experiment")
def test_get_run_attributes() -> None:
    name = "test-dataset"
    ds = new_data_set(name)
    ds.mark_started()
    ds.mark_completed()
    ds.add_metadata("foo", "bar")

    loaded_raw_attrs = get_raw_run_attributes(ds.conn, ds.guid)
    assert loaded_raw_attrs is not None

    assert loaded_raw_attrs["run_id"] == ds.run_id
    assert loaded_raw_attrs["counter"] == ds.counter
    assert loaded_raw_attrs["captured_counter"] == ds.captured_counter
    assert loaded_raw_attrs["captured_run_id"] == ds.captured_run_id
    assert loaded_raw_attrs["captured_run_id"] == ds.captured_run_id
    assert loaded_raw_attrs["experiment"] == get_experiment_attributes_by_exp_id(
        ds.conn, ds.exp_id
    )
    assert loaded_raw_attrs["experiment"]["exp_id"] == ds.exp_id
    assert loaded_raw_attrs["experiment"]["name"] == ds.exp_name
    assert loaded_raw_attrs["experiment"]["sample_name"] == ds.sample_name
    assert loaded_raw_attrs["name"] == name
    assert loaded_raw_attrs["run_timestamp"] == ds.run_timestamp_raw
    assert loaded_raw_attrs["completed_timestamp"] == ds.completed_timestamp_raw
    assert loaded_raw_attrs["parent_dataset_links"] == "[]"
    assert "interdependencies" in loaded_raw_attrs["run_description"]
    assert loaded_raw_attrs["snapshot"] is None
    assert loaded_raw_attrs["metadata"] == {"foo": "bar"}

    loaded_attrs = get_run_attributes(ds.conn, ds.guid)
    assert loaded_attrs is not None

    assert loaded_attrs["run_id"] == ds.run_id
    assert loaded_attrs["counter"] == ds.counter
    assert loaded_attrs["captured_counter"] == ds.captured_counter
    assert loaded_attrs["captured_run_id"] == ds.captured_run_id
    assert loaded_attrs["captured_run_id"] == ds.captured_run_id
    assert loaded_attrs["experiment"] == get_experiment_attributes_by_exp_id(
        ds.conn, ds.exp_id
    )
    assert loaded_attrs["experiment"]["exp_id"] == ds.exp_id
    assert loaded_attrs["name"] == name
    assert loaded_attrs["run_timestamp"] == raw_time_to_str_time(ds.run_timestamp_raw)
    assert loaded_attrs["completed_timestamp"] == raw_time_to_str_time(
        ds.completed_timestamp_raw
    )
    assert loaded_attrs["parent_dataset_links"] == []
    assert isinstance(loaded_attrs["run_description"], RunDescriber)
    assert loaded_attrs["snapshot"] is None
    assert loaded_attrs["metadata"] == {"foo": "bar"}


@pytest.mark.usefixtures("empty_temp_db")
def test_experiment_info_in_dataset():
    exp = new_experiment(name="for_loading", sample_name="no_sample")
    ds = new_data_set("my_first_ds")

    assert ds.exp_id == exp.exp_id
    assert ds.exp_name == exp.name
    assert ds.sample_name == exp.sample_name


@pytest.mark.usefixtures("empty_temp_db")
def test_run_timestamp():
    _ = new_experiment(name="for_loading", sample_name="no_sample")

    t_before_data_set = time.time()
    ds = new_data_set("my_first_ds")
    ds.mark_started()
    t_after_data_set = time.time()

    actual_run_timestamp_raw = ds.run_timestamp_raw

    assert t_before_data_set <= actual_run_timestamp_raw <= t_after_data_set


@pytest.mark.usefixtures("empty_temp_db")
def test_run_timestamp_with_default_format():
    _ = new_experiment(name="for_loading", sample_name="no_sample")

    t_before_data_set = time.time()
    ds = new_data_set("my_first_ds")
    ds.mark_started()
    t_after_data_set = time.time()

    # Note that here we also test the default format of `run_timestamp`
    actual_run_timestamp_raw = time.mktime(
        time.strptime(ds.run_timestamp(), "%Y-%m-%d %H:%M:%S"))

    # Note that because the default format precision is 1 second, we add this
    # second to the right side of the comparison
    t_before_data_set_secs = floor(t_before_data_set)
    t_after_data_set_secs = floor(t_after_data_set)
    assert t_before_data_set_secs \
           <= actual_run_timestamp_raw \
           <= t_after_data_set_secs + 1


@pytest.mark.usefixtures("empty_temp_db")
def test_completed_timestamp():
    _ = new_experiment(name="for_loading", sample_name="no_sample")
    ds = new_data_set("my_first_ds")

    t_before_complete = time.time()
    ds.mark_started()
    ds.mark_completed()
    t_after_complete = time.time()

    actual_completed_timestamp_raw = ds.completed_timestamp_raw

    assert t_before_complete \
           <= actual_completed_timestamp_raw \
           <= t_after_complete


@pytest.mark.usefixtures("empty_temp_db")
def test_completed_timestamp_for_not_completed_dataset():
    _ = new_experiment(name="for_loading", sample_name="no_sample")
    ds = new_data_set("my_first_ds")

    assert ds.pristine is True
    assert ds.started is False
    assert ds.running is False
    assert ds.completed is False

    assert ds.completed_timestamp_raw is None

    assert ds.completed_timestamp() is None


@pytest.mark.usefixtures("empty_temp_db")
def test_completed_timestamp_with_default_format():
    _ = new_experiment(name="for_loading", sample_name="no_sample")
    ds = new_data_set("my_first_ds")

    t_before_complete = time.time()
    ds.mark_started()
    ds.mark_completed()
    t_after_complete = time.time()

    # Note that here we also test the default format of `completed_timestamp`
    actual_completed_timestamp_raw = time.mktime(
        time.strptime(ds.completed_timestamp(), "%Y-%m-%d %H:%M:%S"))

    # Note that because the default format precision is 1 second, we add this
    # second to the right side of the comparison
    t_before_complete_secs = floor(t_before_complete)
    t_after_complete_secs = floor(t_after_complete)
    assert t_before_complete_secs \
           <= actual_completed_timestamp_raw \
           <= t_after_complete_secs + 1


@pytest.mark.usefixtures('experiment')
def test_load_by_guid(some_interdeps):
    ds = DataSet()
    ds.set_interdependencies(some_interdeps[1])
    ds.mark_started()
    ds.add_results([{'ps1': 1, 'ps2': 2}])

    loaded_ds = load_by_guid(ds.guid)

    assert loaded_ds.the_same_dataset_as(ds)


def test_load_by_run_spec(empty_temp_db, some_interdeps):

    def create_ds_with_exp_id(exp_id):
        ds = DataSet(exp_id=exp_id)
        ds.set_interdependencies(some_interdeps[1])
        ds.mark_started()
        ds.add_results([{'ps1': 1, 'ps2': 2}])
        return ds
    # create 3 experiments that mix two experiment names and two sample names
    exp_names = ["te1", "te2", "te1"]
    sample_names = ["ts1", "ts2", "ts2"]

    exps = [new_experiment(exp_name, sample_name=sample_name)
            for exp_name, sample_name in zip(exp_names, sample_names)]

    created_ds = [create_ds_with_exp_id(exp.exp_id) for exp in exps]

    conn = created_ds[0].conn

    guids = get_guids_from_run_spec(conn=conn)
    assert len(guids) == 3

    # since we are not copying runs from multiple dbs we can always load by
    # captured_run_id and this is equivalent to load_by_id
    for i in range(1, 4):
        loaded_ds = load_by_run_spec(captured_run_id=i,
                                     conn=conn)
        assert loaded_ds.guid == guids[i-1]
        assert loaded_ds.the_same_dataset_as(created_ds[i-1])

    # All the datasets datasets have the same captured counter
    # so we cannot load by that alone
    guids_cc1 = get_guids_from_run_spec(captured_counter=1, conn=conn)
    assert len(guids_cc1) == 3
    with pytest.raises(NameError, match="More than one matching"):
        load_by_run_spec(captured_counter=1)

    # there are two different experiments with exp name "test-experiment1"
    # and thus 2 different datasets with counter=1 and that exp name
    guids_cc1_te1 = get_guids_from_run_spec(captured_counter=1,
                                            experiment_name='te1',
                                            conn=conn)
    assert len(guids_cc1_te1) == 2
    with pytest.raises(NameError, match="More than one matching"):
        load_by_run_spec(captured_counter=1, experiment_name="te1", conn=conn)

    # but for "test-experiment2" there is only one
    guids_cc1_te2 = get_guids_from_run_spec(captured_counter=1,
                                            experiment_name='te2',
                                            conn=conn)
    assert len(guids_cc1_te2) == 1
    loaded_ds = load_by_run_spec(captured_counter=1,
                                 experiment_name="te2",
                                 conn=conn)
    assert loaded_ds.guid == guids_cc1_te2[0]
    assert loaded_ds.the_same_dataset_as(created_ds[1])

    # there are two different experiments with sample name "test_sample2" but
    # different exp names so the counter is not unique
    guids_cc1_ts2 = get_guids_from_run_spec(captured_counter=1,
                                            sample_name='ts2',
                                            conn=conn)
    assert len(guids_cc1_ts2) == 2
    with pytest.raises(NameError, match="More than one matching"):
        load_by_run_spec(captured_counter=1,
                         sample_name="ts2",
                         conn=conn)

    # but for  "test_sample1" there is only one
    guids_cc1_ts1 = get_guids_from_run_spec(captured_counter=1,
                                            sample_name='ts1',
                                            conn=conn)
    assert len(guids_cc1_ts1) == 1
    loaded_ds = load_by_run_spec(captured_counter=1,
                                 sample_name="ts1",
                                 conn=conn)
    assert loaded_ds.the_same_dataset_as(created_ds[0])
    assert loaded_ds.guid == guids_cc1_ts1[0]

    # we can load all 3 if we are specific.
    for i in range(3):
        loaded_ds = load_by_run_spec(captured_counter=1,
                                     experiment_name=exp_names[i],
                                     sample_name=sample_names[i],
                                     conn=conn)
        assert loaded_ds.the_same_dataset_as(created_ds[i])
        assert loaded_ds.guid == guids[i]

    # load a non-existing run
    with pytest.raises(NameError, match="No run matching"):
        load_by_run_spec(captured_counter=10000, sample_name="ts2", conn=conn)

    empty_guid_list = get_guids_from_run_spec(conn=conn,
                                              experiment_name='nosuchexp')
    assert empty_guid_list == []



# This is a minimal test that simply ensures that the object exists,
# basically just to notify us if we do API changes

import qcodes as qc


def test_settings_exist():
    limits = qc.SQLiteSettings.limits
    settings = qc.SQLiteSettings.settings

    assert isinstance(limits, dict)
    assert isinstance(settings, dict)
    assert len(limits) == 10



from typing import Optional, Sequence
from datetime import datetime

from qcodes.instrument.parameter import Parameter
from qcodes.loops import Loop
from qcodes.actions import _actions_snapshot
from qcodes.utils.helpers import full_class
from qcodes.utils.metadata import Metadatable


class Measure(Metadatable):
    """
    Create a DataSet from a single (non-looped) set of actions.

    Args:
        *actions (Any): sequence of actions to perform. Any action that is
            valid in a ``Loop`` can be used here. If an action is a gettable
            ``Parameter``, its output will be included in the DataSet.
            Scalars returned by an action will be saved as length-1 arrays,
            with a dummy setpoint for consistency with other DataSets.
    """
    dummy_parameter = Parameter(name='single',
                                label='Single Measurement',
                                set_cmd=None, get_cmd=None)

    def __init__(self, *actions):
        super().__init__()
        self._dummyLoop = Loop(self.dummy_parameter[0]).each(*actions)

    def run_temp(self, **kwargs):
        """
        Wrapper to run this measurement as a temporary data set
        """
        return self.run(quiet=True, location=False, **kwargs)

    def get_data_set(self, *args, **kwargs):
        return self._dummyLoop.get_data_set(*args, **kwargs)

    def run(self, use_threads=False, quiet=False, station=None, **kwargs):
        """
        Run the actions in this measurement and return their data as a DataSet

        Args:
            quiet (Optional[bool]): Set True to not print anything except
                errors. Default False.

            station (Optional[Station]): the ``Station`` this measurement
                pertains to. Defaults to ``Station.default`` if one is defined.
                Only used to supply metadata.

            use_threads (Optional[bool]): whether to parallelize ``get``
                operations using threads. Default False.

            location (Optional[Union[str, bool]]): the location of the
                DataSet, a string whose meaning depends on formatter and io,
                or False to only keep in memory. May be a callable to provide
                automatic locations. If omitted, will use the default
                DataSet.location_provider

            name (Optional[str]): if location is default or another provider
                function, name is a string to add to location to make it more
                readable/meaningful to users

            formatter (Optional[Formatter]): knows how to read and write the
                file format. Default can be set in DataSet.default_formatter

            io (Optional[io_manager]): knows how to connect to the storage
                (disk vs cloud etc)

        location, name formatter and io are passed to ``data_set.new_data``
        along with any other optional keyword arguments.

        returns:
            a DataSet object containing the results of the measurement
        """

        data_set = self._dummyLoop.get_data_set(**kwargs)

        # set the DataSet to local for now so we don't save it, since
        # we're going to massage it afterward
        original_location = data_set.location
        data_set.location = False

        # run the measurement as if it were a Loop
        self._dummyLoop.run(use_threads=use_threads,
                            station=station, quiet=True)

        # look for arrays that are unnecessarily nested, and un-nest them
        all_unnested = True
        for array in data_set.arrays.values():
            if array.ndim == 1:
                if array.is_setpoint:
                    dummy_setpoint = array
                else:
                    # we've found a scalar - so keep the dummy setpoint
                    all_unnested = False
            else:
                # The original return was an array, so take off the extra dim.
                # (This ensures the outer dim length was 1, otherwise this
                # will raise a ValueError.)
                array.ndarray.shape = array.ndarray.shape[1:]

                # TODO: DataArray.shape masks ndarray.shape, and a user *could*
                # change it, thinking they were reshaping the underlying array,
                # but this would a) not actually reach the ndarray right now,
                # and b) if it *did* and the array was reshaped, this array
                # would be out of sync with its setpoint arrays, so bad things
                # would happen. So we probably want some safeguards in place
                # against this
                array.shape = array.ndarray.shape

                array.set_arrays = array.set_arrays[1:]

                array.init_data()

        # Do we still need the dummy setpoint array at all?
        if all_unnested:
            del data_set.arrays[dummy_setpoint.array_id]
            if hasattr(data_set, 'action_id_map'):
                del data_set.action_id_map[dummy_setpoint.action_indices]

        # now put back in the DataSet location and save it
        data_set.location = original_location
        data_set.write()

        # metadata: ActiveLoop already provides station snapshot, but also
        # puts in a 'loop' section that we need to replace with 'measurement'
        # but we use the info from 'loop' to ensure consistency and avoid
        # duplication.
        LOOP_SNAPSHOT_KEYS = ['ts_start', 'ts_end', 'use_threads']
        data_set.add_metadata({'measurement': {
            k: data_set.metadata['loop'][k] for k in LOOP_SNAPSHOT_KEYS
        }})
        del data_set.metadata['loop']

        # actions are included in self.snapshot() rather than in
        # LOOP_SNAPSHOT_KEYS because they are useful if someone just
        # wants a local snapshot of the Measure object
        data_set.add_metadata({'measurement': self.snapshot()})

        data_set.save_metadata()

        if not quiet:
            print(repr(data_set))
            print(datetime.now().strftime('acquired at %Y-%m-%d %H:%M:%S'))

        return data_set

    def snapshot_base(self, update: Optional[bool] = False,
                      params_to_skip_update: Optional[Sequence[str]] = None):
        return {
            '__class__': full_class(self),
            'actions': _actions_snapshot(self._dummyLoop.actions, update)
        }



import importlib
from typing import TYPE_CHECKING, Any, List, Mapping

from qcodes.instrument.base import InstrumentBase

if TYPE_CHECKING:
    from qcodes.station import Station


class InstrumentGroup(InstrumentBase):
    """
    InstrumentGroup is an instrument driver to represent a series of instruments
    that are grouped together. This instrument is mainly used as a wrapper for
    sub instruments/submodules and particularly built for use with grouping
    multiple :class:`DelegateInstrument` s.

    Args:
        name: Name referring to this group of items
        station: Measurement station with real instruments
        submodules_type: Class to use for creating the instruments.
        submodules: A mapping between an instrument name and the values passed
            to the constructor of the class specified by `submodules_type`.
        initial_values: A mapping between the names of parameters and initial
            values to set on those parameters when loading this instrument.
        set_initial_values_on_load: Set default values on load. Defaults to
            False.
    """
    def __init__(
        self,
        name: str,
        station: "Station",
        submodules_type: str,
        submodules: Mapping[str, Mapping[str, List[str]]],
        initial_values: Mapping[str, Mapping[str, Any]],
        set_initial_values_on_load: bool = False,
        **kwargs: Any
    ):
        super().__init__(name=name, **kwargs)

        module_name = '.'.join(submodules_type.split('.')[:-1])
        instr_class_name = submodules_type.split('.')[-1]
        module = importlib.import_module(module_name)
        instr_class = getattr(module, instr_class_name)

        for submodule_name, inputs in submodules.items():
            if not any(x in inputs.keys() for x in ["parameters", "channels"]):
                raise KeyError(
                    f"Missing keyworded input arguments for {submodule_name}"
                )
            submodule = instr_class(
                name=submodule_name,
                station=station,
                **inputs,
                initial_values=initial_values.get(submodule_name),
                set_initial_values_on_load=set_initial_values_on_load
            )

            self.add_submodule(
                submodule_name,
                submodule
            )

    def __repr__(self) -> str:
        submodules = ", ".join(self.submodules.keys())
        return f"InstrumentGroup(name={self.name}, submodules={submodules})"



r"""
This module encapsulates QCoDeS database: its schema, structure, convenient
and relevant queries, wrapping around :mod:`sqlite3`, etc.

The dependency structure of the sub-modules is the following:

::

         .connection     .settings
           /   |   \         |
          /    |    \        |
         /     |     V       V
        |      |  .query_helpers
        |      |   |       |
        |      V   V       |
        |  .db_upgrades    |
        |      /           V
        |     /        .queries
        v    v
      .database

"""



import warnings
import pytest


from qcodes.utils.deprecate import (
    deprecate, issue_deprecation_warning, _catch_deprecation_warnings,
    assert_not_deprecated, assert_deprecated, QCoDeSDeprecationWarning)


def test_assert_deprecated_raises():
    with assert_deprecated(
            'The use of this function is deprecated, because '
            'of this being a test. Use \"a real function\" as an '
            'alternative.'):
        issue_deprecation_warning(
            'use of this function',
            'of this being a test',
            'a real function'
        )


def test_assert_deprecated_does_not_raise_wrong_msg():
    with pytest.raises(AssertionError):
        with assert_deprecated('entirely different message'):
            issue_deprecation_warning('warning')


def test_assert_deprecated_does_not_raise_wrong_type():
    with pytest.raises(AssertionError):
        with assert_deprecated('entirely different message'):
            warnings.warn('warning')


def test_assert_not_deprecated_raises():
    with pytest.raises(AssertionError):
        with assert_not_deprecated():
            issue_deprecation_warning('something')


def test_assert_not_deprecated_does_not_raise():
    with assert_not_deprecated():
        warnings.warn('some other warning')


@pytest.mark.filterwarnings('ignore:The function "add_one" is deprecated,')
def test_similar_output():

    def _add_one(x):
        return 1 + x

    @deprecate(reason='this function is for private use only')
    def add_one(x):
        return _add_one(x)

    with assert_deprecated(
            'The function <add_one> is deprecated, because '
            'this function is for private use only.'):
        assert add_one(1) == _add_one(1)


def test_deprecated_context_manager():
    with _catch_deprecation_warnings() as ws:
        issue_deprecation_warning('something')
        issue_deprecation_warning('something more')
        warnings.warn('Some other warning')
    assert len(ws) == 2

    with pytest.warns(expected_warning=QCoDeSDeprecationWarning) as ws:
        issue_deprecation_warning('something')
        warnings.warn('Some other warning')
    assert len(ws) == 2


@deprecate(reason='this is a test')
class C:
    def __init__(self, attr: str):
        self.a = attr

    def method(self) -> None:
        self.a = 'last called by method'

    @staticmethod
    def static_method(a: int) -> int:
        return a + 1

    @classmethod
    def class_method(cls, a: int) -> int:
        return a + 1

    @property
    def prop(self) -> str:
        return self.a

    @prop.setter
    def prop(self, val: str) -> None:
        self.a = val + '_prop'


def test_init_uninhibited():
    with pytest.warns(expected_warning=QCoDeSDeprecationWarning):
        c = C('pristine')
    assert c.a == 'pristine'


def test_init_raises():
    with assert_deprecated(
            'The class <C> is deprecated, because '
            'this is a test.'):
        C('pristine')


def test_method_uninhibited():
    with pytest.warns(expected_warning=QCoDeSDeprecationWarning):
        c = C('pristine')
        c.method()
    assert c.a == 'last called by method'


def test_method_raises():
    with pytest.warns(expected_warning=QCoDeSDeprecationWarning):
        c = C('pristine')

    with assert_deprecated(
            'The function <method> is deprecated, because '
            'this is a test.'):
        c.method()


def test_static_method_uninhibited():
    assert C.static_method(1) == 2
    with pytest.warns(expected_warning=QCoDeSDeprecationWarning):
        c = C('pristine')
        assert c.static_method(1) == 2


@pytest.mark.xfail(reason="This is not implemented yet.")
def test_static_method_raises(self):
    with assert_deprecated(
            'The function <static_method> is deprecated, because '
            'this is a test.'):
        assert C.static_method(1) == 2
    with pytest.warns(expected_warning=QCoDeSDeprecationWarning):
        c = C('pristine')
    with assert_deprecated(
            'The function <static_method> is deprecated, because '
            'this is a test.'):
        assert c.static_method(1) == 2


def test_class_method_uninhibited():
    with pytest.warns(expected_warning=QCoDeSDeprecationWarning):
        assert C.class_method(1) == 2
    with pytest.warns(expected_warning=QCoDeSDeprecationWarning):
        c = C('pristine')
    with pytest.warns(expected_warning=QCoDeSDeprecationWarning):
        assert c.class_method(1) == 2


@pytest.mark.xfail(reason="This is not implemented yet.")
def test_class_method_raises():
    with assert_deprecated(
            'The function <static_method> is deprecated, because '
            'this is a test.'):
        C.class_method(1)
    with pytest.warns(expected_warning=QCoDeSDeprecationWarning):
        c = C('pristine')
    with assert_deprecated(
            'The function <static_method> is deprecated, because '
            'this is a test.'):
        c.class_method(1)


def test_property_uninhibited():
    with pytest.warns(expected_warning=QCoDeSDeprecationWarning):
        c = C('pristine')
        assert c.prop == 'pristine'


@pytest.mark.xfail(reason="This is not implemented yet.")
def test_property_raises():
    with pytest.warns(expected_warning=QCoDeSDeprecationWarning):
        c = C('pristine')
        with assert_deprecated(
                'The function <method> is deprecated, because '
                'this is a test.'):
            _ = c.prop  # pylint: disable=pointless-statement


def test_setter_uninhibited():
    with pytest.warns(expected_warning=QCoDeSDeprecationWarning):
        c = C('pristine')
        c.prop = 'changed'
        assert c.a == 'changed_prop'


@pytest.mark.xfail(reason="This is not implemented yet.")
def test_setter_raises(self):
    with pytest.warns(expected_warning=QCoDeSDeprecationWarning):
        c = C('pristine')

    with assert_deprecated(
            'The function <method> is deprecated, because '
            'this is a test.'):
        c.prop = 'changed'



"""
Tests for plotting system.
Legacy in many ways:

    - assume X server running
    - just test "window creation"
"""
from unittest import TestCase, skipIf
import numpy as np
import os

try:
    noQtPlot = False
    from qcodes.plots.pyqtgraph import QtPlot
except Exception:
    noQtPlot = True

try:
    noMatPlot = False
    from qcodes.plots.qcmatplotlib import MatPlot
    import matplotlib
    matplotlib.use('Agg')
    import matplotlib.pyplot as plt
except Exception:
    noMatPlot = True


@skipIf(noQtPlot, '***pyqtgraph plotting cannot be tested***')
class TestQtPlot(TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_creation(self):
        """
        Simple test function which created a QtPlot window
        """
        plotQ = QtPlot(remote=False, show_window=False, interval=0)
        plotQ.add_subplot()

    def test_simple_plot(self):
        main_QtPlot = QtPlot(
            window_title='Main plotmon of TestQtPlot',
            figsize=(600, 400))

        x = np.arange(0, 10e-6, 1e-9)
        f = 2e6
        y = np.cos(2*np.pi*f*x)

        for j in range(4):
            main_QtPlot.add(x=x, y=y,
                            xlabel='Time', xunit='s',
                            ylabel='Amplitude', yunit='V',
                            subplot=j+1,
                            symbol='o', symbolSize=5)

    def test_return_handle(self):
        plotQ = QtPlot(remote=False)
        return_handle = plotQ.add([1, 2, 3])
        self.assertIs(return_handle, plotQ.subplots[0].items[0])


@skipIf(noMatPlot, '***matplotlib plotting cannot be tested***')
class TestMatPlot(TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_creation(self):
        """
        Simple test function which created a MatPlot window
        """
        plotM = MatPlot(interval=0)
        plt.close(plotM.fig)

    def test_return_handle(self):
        plotM = MatPlot(interval=0)
        returned_handle = plotM.add([1, 2, 3])
        line_handle = plotM[0].get_lines()[0]
        self.assertIs(returned_handle, line_handle)
        plotM.clear()
        plt.close(plotM.fig)



from typing import Any

import numpy as np
from qcodes import (ChannelList, InstrumentChannel, ParameterWithSetpoints,
                    VisaInstrument)
from qcodes.utils.validators import Arrays, Enum, Numbers


class RigolDS1074ZChannel(InstrumentChannel):
    """
    Contains methods and attributes specific to the Rigol
    oscilloscope channels.

    The output trace from each channel of the oscilloscope
    can be obtained using 'trace' parameter.
    """

    def __init__(self,
                 parent: "DS1074Z",
                 name: str,
                 channel: int
                 ):
        super().__init__(parent, name)
        self.channel = channel

        self.add_parameter("vertical_scale",
                           get_cmd=f":CHANnel{channel}:SCALe?",
                           set_cmd=":CHANnel{}:SCALe {}".format(channel, "{}"),
                           get_parser=float
                           )

        self.add_parameter("trace",
                           get_cmd=self._get_full_trace,
                           vals=Arrays(shape=(self.parent.waveform_npoints,)),
                           setpoints=(self.parent.time_axis,),
                           unit='V',
                           parameter_class=ParameterWithSetpoints,
                           snapshot_value=False
                           )

    def _get_full_trace(self) -> np.ndarray:
        y_ori = self.root_instrument.waveform_yorigin()
        y_increm = self.root_instrument.waveform_yincrem()
        y_ref = self.root_instrument.waveform_yref()
        y_raw = self._get_raw_trace()
        y_raw_shifted = y_raw - y_ori - y_ref
        full_data = np.multiply(y_raw_shifted, y_increm)
        return full_data

    def _get_raw_trace(self) -> np.ndarray:
        # set the out type from oscilloscope channels to WORD
        self.root_instrument.write(':WAVeform:FORMat WORD')

        # set the channel from where data will be obtained
        self.root_instrument.data_source(f"ch{self.channel}")

        # Obtain the trace
        raw_trace_val = self.root_instrument.visa_handle.query_binary_values(
            'WAV:DATA?',
            datatype='h',
            is_big_endian=False,
            expect_termination=False)
        return np.array(raw_trace_val)


class DS1074Z(VisaInstrument):
    """
    The QCoDeS drivers for Oscilloscope Rigol DS1074Z.

    Args:
        name: name of the instrument.
        address: VISA address of the instrument.
        timeout: Seconds to allow for responses.
        terminator: terminator for SCPI commands.
    """
    def __init__(
            self,
            name: str,
            address: str,
            terminator: str = '\n',
            timeout: float = 5,
            **kwargs: Any):
        super().__init__(name, address, terminator=terminator, timeout=timeout,
                         **kwargs)

        self.add_parameter('waveform_xorigin',
                           get_cmd='WAVeform:XORigin?',
                           unit='s',
                           get_parser=float
                           )

        self.add_parameter('waveform_xincrem',
                           get_cmd=':WAVeform:XINCrement?',
                           unit='s',
                           get_parser=float
                           )

        self.add_parameter('waveform_npoints',
                           get_cmd='WAV:POIN?',
                           set_cmd='WAV:POIN {}',
                           unit='s',
                           get_parser=int
                           )

        self.add_parameter('waveform_yorigin',
                           get_cmd='WAVeform:YORigin?',
                           unit='V',
                           get_parser=float
                           )

        self.add_parameter('waveform_yincrem',
                           get_cmd=':WAVeform:YINCrement?',
                           unit='V',
                           get_parser=float
                           )

        self.add_parameter('waveform_yref',
                           get_cmd=':WAVeform:YREFerence?',
                           unit='V',
                           get_parser=float
                           )

        self.add_parameter('trigger_mode',
                           get_cmd=':TRIGger:MODE?',
                           set_cmd=':TRIGger:MODE {}',
                           unit='V',
                           vals=Enum('edge',
                                     'pulse',
                                     'video',
                                     'pattern'
                                     ),
                           get_parser=str
                           )

        # trigger source
        self.add_parameter('trigger_level',
                           unit='V',
                           get_cmd=self._get_trigger_level,
                           set_cmd=self._set_trigger_level,
                           vals=Numbers()
                           )

        self.add_parameter('trigger_edge_source',
                           label='Source channel for the edge trigger',
                           get_cmd=':TRIGger:EDGE:SOURce?',
                           set_cmd=':TRIGger:EDGE:SOURce {}',
                           val_mapping={'ch1': 'CHAN1',
                                        'ch2': 'CHAN2',
                                        'ch3': 'CHAN3',
                                        'ch4': 'CHAN4'
                                        }
                           )

        self.add_parameter('trigger_edge_slope',
                           label='Slope of the edge trigger',
                           get_cmd=':TRIGger:EDGE:SLOPe?',
                           set_cmd=':TRIGger:EDGE:SLOPe {}',
                           vals=Enum('positive', 'negative', 'neither')
                           )

        self.add_parameter('data_source',
                           label='Waveform Data source',
                           get_cmd=':WAVeform:SOURce?',
                           set_cmd=':WAVeform:SOURce {}',
                           val_mapping={'ch1': 'CHAN1',
                                        'ch2': 'CHAN2',
                                        'ch3': 'CHAN3',
                                        'ch4': 'CHAN4'
                                        }
                           )

        self.add_parameter('time_axis',
                           unit='s',
                           label='Time',
                           set_cmd=False,
                           get_cmd=self._get_time_axis,
                           vals=Arrays(shape=(self.waveform_npoints,)),
                           snapshot_value=False
                           )

        channels = ChannelList(self,
                               "channels",
                               RigolDS1074ZChannel,
                               snapshotable=False
                               )

        for channel_number in range(1, 5):
            channel = RigolDS1074ZChannel(self,
                                          f"ch{channel_number}",
                                          channel_number
                                          )
            channels.append(channel)

        channels.lock()
        self.add_submodule('channels', channels)

        self.connect_message()

    def _get_time_axis(self) -> np.ndarray:
        xorigin = self.waveform_xorigin()
        xincrem = self.waveform_xincrem()
        npts = self.waveform_npoints()
        xdata = np.linspace(xorigin, npts * xincrem + xorigin, npts)
        return xdata

    def _get_trigger_level(self) -> str:
        trigger_level = self.root_instrument.ask(f":TRIGger:{self.trigger_mode()}:LEVel?")
        return trigger_level

    def _set_trigger_level(self, value: str) -> None:
        self.root_instrument.write(f":TRIGger:{self.trigger_mode()}:LEVel {value}")



import pytest

import numpy as np

from qcodes.instrument.parameter import Parameter
import qcodes.utils.validators as vals
from .conftest import BookkeepingValidator


def test_number_of_validations():
    p = Parameter('p', set_cmd=None, initial_value=0,
                  vals=BookkeepingValidator())
    # in the set wrapper the final value is validated
    # and then subsequently each step is validated.
    # in this case there is one step so the final value
    # is validated twice.
    assert p.vals.values_validated == [0, 0]

    p.step = 1
    p.set(10)
    assert p.vals.values_validated == [0, 0, 10, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10]


def test_number_of_validations_for_set_cache():
    p = Parameter('p', set_cmd=None,
                  vals=BookkeepingValidator())
    assert p.vals.values_validated == []

    p.cache.set(1)
    assert p.vals.values_validated == [1]

    p.cache.set(4)
    assert p.vals.values_validated == [1, 4]

    p.step = 1
    p.cache.set(10)
    assert p.vals.values_validated == [1, 4, 10]


def test_bad_validator():
    with pytest.raises(TypeError):
        Parameter('p', vals=[1, 2, 3])


def test_setting_int_with_float():
    parameter = Parameter(name='foobar', set_cmd=None, get_cmd=None,
                          set_parser=lambda x: int(round(x)),
                          vals=vals.PermissiveInts(0))

    a = 0
    b = 10
    values = np.linspace(a, b, b-a+1)
    for i in values:
        parameter(i)
        a = parameter()
        assert isinstance(a, int)


def test_setting_int_with_float_not_close():
    parameter = Parameter(name='foobar', set_cmd=None, get_cmd=None,
                          set_parser=lambda x: int(round(x)),
                          vals=vals.PermissiveInts(0))

    a = 0
    b = 10
    values = np.linspace(a, b, b-a+2)
    for i in values[1:-2]:
        with pytest.raises(TypeError):
            parameter(i)



import json
import logging
import sys
from collections import defaultdict
from typing import DefaultDict, Dict, List, Mapping, Sequence, Tuple

from tqdm import tqdm

from qcodes.dataset.descriptions.param_spec import ParamSpec
from qcodes.dataset.descriptions.versioning.v0 import InterDependencies
from qcodes.dataset.sqlite.connection import (
    ConnectionPlus,
    atomic,
    atomic_transaction,
    transaction,
)
from qcodes.dataset.sqlite.query_helpers import one

log = logging.getLogger(__name__)


def _2to3_get_result_tables(conn: ConnectionPlus) -> Dict[int, str]:
    rst_query = "SELECT run_id, result_table_name FROM runs"
    cur = conn.cursor()
    cur.execute(rst_query)

    data = cur.fetchall()
    cur.close()
    results = {}
    for row in data:
        results[row['run_id']] = row['result_table_name']
    return results


def _2to3_get_layout_ids(conn: ConnectionPlus) -> DefaultDict[int, List[int]]:
    query = """
            select runs.run_id, layouts.layout_id
            FROM layouts
            INNER JOIN runs ON runs.run_id == layouts.run_id
            """
    cur = conn.cursor()
    cur.execute(query)
    data = cur.fetchall()
    cur.close()

    results: DefaultDict[int, List[int]] = defaultdict(list)

    for row in data:
        run_id = row['run_id']
        layout_id = row['layout_id']
        results[run_id].append(layout_id)

    return results


def _2to3_get_indeps(conn: ConnectionPlus) -> DefaultDict[int, List[int]]:
    query = """
            SELECT layouts.run_id, layouts.layout_id
            FROM layouts
            INNER JOIN dependencies
            ON layouts.layout_id==dependencies.independent
            """
    cur = conn.cursor()
    cur.execute(query)
    data = cur.fetchall()
    cur.close()
    results: DefaultDict[int, List[int]] = defaultdict(list)

    for row in data:
        run_id = row['run_id']
        layout_id = row['layout_id']
        results[run_id].append(layout_id)

    return results


def _2to3_get_deps(conn: ConnectionPlus) -> DefaultDict[int, List[int]]:
    query = """
            SELECT layouts.run_id, layouts.layout_id
            FROM layouts
            INNER JOIN dependencies
            ON layouts.layout_id==dependencies.dependent
            """
    cur = conn.cursor()
    cur.execute(query)
    data = cur.fetchall()
    cur.close()
    results: DefaultDict[int, List[int]] = defaultdict(list)

    for row in data:
        run_id = row['run_id']
        layout_id = row['layout_id']
        results[run_id].append(layout_id)

    return results


def _2to3_get_dependencies(conn: ConnectionPlus
                           ) -> DefaultDict[int, List[int]]:
    query = """
            SELECT dependent, independent
            FROM dependencies
            ORDER BY dependent, axis_num ASC
            """
    cur = conn.cursor()
    cur.execute(query)
    data = cur.fetchall()
    cur.close()
    results: DefaultDict[int, List[int]] = defaultdict(list)

    if len(data) == 0:
        return results

    for row in data:
        dep = row['dependent']
        indep = row['independent']
        results[dep].append(indep)

    return results


def _2to3_get_layouts(conn: ConnectionPlus) -> Dict[int,
                                                    Tuple[str, str, str, str]]:
    query = """
            SELECT layout_id, parameter, label, unit, inferred_from
            FROM layouts
            """
    cur = conn.cursor()
    cur.execute(query)

    results: Dict[int, Tuple[str, str, str, str]] = {}
    for row in cur.fetchall():
        results[row['layout_id']] = (row['parameter'],
                                     row['label'],
                                     row['unit'],
                                     row['inferred_from'])
    return results


def _2to3_get_paramspecs(
    conn: ConnectionPlus,
    layout_ids: List[int],
    layouts: Mapping[int, Tuple[str, str, str, str]],
    dependencies: Mapping[int, Sequence[int]],
    deps: Sequence[int],
    indeps: Sequence[int],
    result_table_name: str,
) -> Dict[int, ParamSpec]:

    paramspecs: Dict[int, ParamSpec] = {}

    the_rest = set(layout_ids).difference(set(deps).union(set(indeps)))

    # We ensure that we first retrieve the ParamSpecs on which other ParamSpecs
    # depend, then the dependent ParamSpecs and finally the rest

    for layout_id in list(indeps) + list(deps) + list(the_rest):
        (name, label, unit, inferred_from_str) = layouts[layout_id]
        # get the data type
        sql = f'PRAGMA TABLE_INFO("{result_table_name}")'
        c = transaction(conn, sql)
        paramtype = None
        for row in c.fetchall():
            if row['name'] == name:
                paramtype = row['type']
                break
        if paramtype is None:
            raise TypeError(f"Could not determine type of {name} during the"
                            f"db upgrade of {result_table_name}")

        inferred_from: List[str] = []
        depends_on: List[str] = []

        # this parameter depends on another parameter
        if layout_id in deps:
            setpoints = dependencies[layout_id]
            depends_on = [paramspecs[idp].name for idp in setpoints]

        if inferred_from_str != '':
            inferred_from = inferred_from_str.split(', ')

        paramspec = ParamSpec(name=name,
                              paramtype=paramtype,
                              label=label, unit=unit,
                              depends_on=depends_on,
                              inferred_from=inferred_from)
        paramspecs[layout_id] = paramspec

    return paramspecs


def upgrade_2_to_3(conn: ConnectionPlus) -> None:
    """
    Perform the upgrade from version 2 to version 3

    Insert a new column, run_description, to the runs table and fill it out
    for exisitng runs with information retrieved from the layouts and
    dependencies tables represented as the json output of a RunDescriber
    object
    """

    no_of_runs_query = "SELECT max(run_id) FROM runs"
    no_of_runs = one(atomic_transaction(conn, no_of_runs_query), 'max(run_id)')
    no_of_runs = no_of_runs or 0

    # If one run fails, we want the whole upgrade to roll back, hence the
    # entire upgrade is one atomic transaction

    with atomic(conn) as conn:
        sql = "ALTER TABLE runs ADD COLUMN run_description TEXT"
        transaction(conn, sql)

        result_tables = _2to3_get_result_tables(conn)
        layout_ids_all = _2to3_get_layout_ids(conn)
        indeps_all = _2to3_get_indeps(conn)
        deps_all = _2to3_get_deps(conn)
        layouts = _2to3_get_layouts(conn)
        dependencies = _2to3_get_dependencies(conn)

        pbar = tqdm(range(1, no_of_runs+1), file=sys.stdout)
        pbar.set_description("Upgrading database; v2 -> v3")

        for run_id in pbar:

            if run_id in layout_ids_all:

                result_table_name = result_tables[run_id]
                layout_ids = list(layout_ids_all[run_id])
                if run_id in indeps_all:
                    independents = tuple(indeps_all[run_id])
                else:
                    independents = ()
                if run_id in deps_all:
                    dependents = tuple(deps_all[run_id])
                else:
                    dependents = ()

                paramspecs = _2to3_get_paramspecs(conn,
                                                  layout_ids,
                                                  layouts,
                                                  dependencies,
                                                  dependents,
                                                  independents,
                                                  result_table_name)

                interdeps = InterDependencies(*paramspecs.values())
                desc_dict = {'interdependencies': interdeps._to_dict()}
                json_str = json.dumps(desc_dict)

            else:
                desc_dict = {'interdependencies':
                                 InterDependencies()._to_dict()}
                json_str = json.dumps(desc_dict)

            sql = f"""
                   UPDATE runs
                   SET run_description = ?
                   WHERE run_id == ?
                   """
            cur = conn.cursor()
            cur.execute(sql, (json_str, run_id))
            log.debug(f"Upgrade in transition, run number {run_id}: OK")



from collections import OrderedDict, namedtuple
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Union,
)

from qcodes.instrument.group_parameter import Group, GroupParameter
from qcodes.instrument.parameter import (
    DelegateParameter,
    ParamDataType,
    Parameter,
    ParamRawDataType,
    _BaseParameter,
)

if TYPE_CHECKING:
    from qcodes.instrument.base import InstrumentBase

import logging

_log = logging.getLogger(__name__)


class DelegateGroupParameter(DelegateParameter, GroupParameter):

    def __init__(
        self,
        name: str,
        source: Optional[Parameter],
        instrument: Optional['InstrumentBase'] = None,
        initial_value: Union[float, str, None] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            name=name,
            source=source,
            instrument=instrument,
            initial_value=initial_value,
            **kwargs,
        )


class DelegateGroup(Group):
    """The DelegateGroup combines :class:`.DelegateParameter` s that
    are to be gotten or set using one :class:`.GroupedParameter`.
    Each :class:`.DelegateParameter` maps to one source parameter
    that is individually set or gotten on an instrument. These
    parameters can originate from the same or different instruments.

    The class :class:`.DelegateGroup` is used within the
    :class:`GroupedParameter` class in order to get and set the
    :class:`.DelegateParameter` s either via their default get and set
    methods or via a custom get or set method.

    The value to be set can be passed to the set method either via a
    dictionary, where the keys are the names of the
    :class:`.DelegateParameter` s contained in the :class:`DelegateGroup`,
    or a single value, if a custom setter is defined or if the group
    only contains a single :class:`.DelegateParameter`.

    The value returned by the get method is passed through a formatter.
    By default, the formatter returns the :class:`.DelegateParameter`
    values in a namedtuple, where the keys are the names of the
    :class:`.DelegateParameter` s. In the special case where the
    :class:`.DelegateGroup` only contains one :class:`.DelegateParameter`,
    the formatter simply returns the individual value. Optionally,
    the formatter can be customized and specified via the constructor.
    The formatter takes as input the values of the :class:`.DelegateParameter` s
    as positional arguments in the order at which the
    :class:`.DelegateParameter` s are specified.

    Args:
        name: Name of the DelegateGroup
        parameters: DelegateParameters to group together
        parameter_names: Optional names of parameters, defaults to the
            parameter `name` attributes
        setter: Optional function to call for setting the grouped parameters,
            should take one argument `value`. Defaults to set_parameters(),
            which sets each parameter using its .set() method.
        getter: Optional function to call for getting the grouped parameters.
            Defaults to .get_parameters(), which runs the get() method for each
            parameter.
        formatter: Optional formatter for value returned by get_parameters(),
            defaults to a namedtuple with the parameter names as keys.
    """
    def __init__(
        self,
        name: str,
        parameters: Sequence[DelegateGroupParameter],
        parameter_names: Optional[Iterable[str]] = None,
        setter: Optional[Callable[..., Any]] = None,
        getter: Optional[Callable[..., Any]] = None,
        formatter: Optional[Callable[..., Any]] = None,
        **kwargs: Any
    ):
        super().__init__(
            parameters=parameters,
            single_instrument=False,
            **kwargs
        )

        self.name = name
        self._parameter_names = parameter_names or [
            _e.name for _e in parameters
        ]
        self._set_fn = setter
        self._get_fn = getter
        self._params = parameters
        self._parameters = OrderedDict(zip(self._parameter_names, parameters))

        if formatter is None and len(parameters) == 1:
            self._formatter = lambda result: result
        elif formatter is None:
            self._formatter = self._namedtuple
        else:
            self._formatter = formatter

    def _namedtuple(self, *args: Any, **kwargs: Any) -> Tuple[Any, ...]:
        return namedtuple(self.name, self._parameter_names)(*args, **kwargs)

    def set(self, value: Union[ParamDataType, Mapping[str, ParamDataType]]) -> None:
        if self._set_fn is not None:
            self._set_fn(value)
        else:
            if not isinstance(value, dict):
                value = {
                    name: value for name in self._parameter_names
                }
            self.set_parameters(value)

    def get(self) -> Any:
        if self._get_fn is not None:
            return self._get_fn()
        else:
            return self.get_parameters()

    def get_parameters(self) -> Any:
        return self._formatter(*(_p.get() for _p in self.parameters.values()))

    def _set_from_dict(self, calling_dict: Mapping[str, ParamRawDataType]) -> None:
        for name, p in list(self.parameters.items()):
            p.set(calling_dict[name])

    @property
    def source_parameters(self) -> Tuple[Optional[Parameter], ...]:
        """Get source parameters of each DelegateParameter"""
        return tuple(p.source for p in self._params)


class GroupedParameter(_BaseParameter):
    """
    The GroupedParameter wraps one or more :class:`.DelegateParameter` s,
    such that those parameters can be accessed as if they were one
    parameter.

    The :class:`GroupedParameter` uses a :class:`DelegateGroup` to keep
    track of the :class:`.DelegateParameter` s. Mainly, this class is a
    thin wrapper around the :class:`DelegateGroup`, and mainly exists
    in order to allow for it to be used as a :class:`_BaseParameter`.

    This class can be seen as the opposite of a :class:`GroupParameter`,
    which is a class to create parameters that are set with a single get
    and set string command on *the same* instrument but need to be accessed
    separately. In contrast, the :class:`GroupedParameter` allows grouped
    access to parameters that are normally separate, and can be associated
    with *different* instruments.

    Args:
        name: Grouped parameter name.
        group: Group that contains the target parameter(s).
        unit: The unit of measure. Use ``''`` for unitless.
        label: Optional label, defaults to parameter name.
        default set method(s).
    """
    def __init__(
        self,
        name: str,
        group: DelegateGroup,
        unit: Optional[str] = None,
        label: Optional[str] = None,
        **kwargs: Any
    ):
        super().__init__(name, **kwargs)
        self.label = name if label is None else label
        self.unit = unit if unit is not None else ''
        self._group = group

    @property
    def group(self) -> 'DelegateGroup':
        """
        The group that contains the target parameters.
        """
        return self._group

    @property
    def parameters(self) -> Dict[str, GroupParameter]:
        """Get delegate parameters wrapped by this GroupedParameter"""
        return self.group.parameters

    @property
    def source_parameters(self) -> Tuple[Optional[Parameter], ...]:
        """Get source parameters of each DelegateParameter"""
        return self.group.source_parameters

    def get_raw(self) -> Union[ParamDataType, Mapping[str, ParamDataType]]:
        """Get parameter raw value"""
        return self.group.get_parameters()

    def set_raw(self, value: Union[ParamDataType, Mapping[str, ParamDataType]]) -> None:
        """Set parameter raw value

        Args:
            value: Parameter value to set

        Returns:
            float: Returns the parameter value
        """
        self.group.set(value)

    def __repr__(self) -> str:
        output = f"GroupedParameter(name={self.name}"
        if self.source_parameters:
            source_parameters = ", ".join(str(_) for _ in self.source_parameters)
            output += f", source_parameters=({source_parameters})"
        output += ")"
        return output



"""Ethernet instrument driver class based on sockets."""
import socket
import logging
from typing import Dict, Sequence, Optional, Any, Type
from types import TracebackType

from .base import Instrument

log = logging.getLogger(__name__)


class IPInstrument(Instrument):
    r"""
    Bare socket ethernet instrument implementation. Use of `VisaInstrument`
    is promoted instead of this.

    Args:
        name: What this instrument is called locally.

        address: The IP address or name. If not given on
            construction, must be provided before any communication.

        port: The IP port. If not given on construction, must
            be provided before any communication.

        timeout: Seconds to allow for responses. Default 5.

        terminator: Character(s) to terminate each send. Default '\n'.

        persistent: Whether to leave the socket open between calls.
            Default True.

        write_confirmation: Whether the instrument acknowledges writes
            with some response we should read. Default True.

        kwargs: additional static metadata to add to this
            instrument's JSON snapshot.

    See help for ``qcodes.Instrument`` for additional information on writing
    instrument subclasses.
    """

    def __init__(self, name: str,
                 address: Optional[str] = None,
                 port: Optional[int] = None,
                 timeout: float = 5,
                 terminator: str = '\n',
                 persistent: bool = True,
                 write_confirmation: bool = True,
                 **kwargs: Any):
        super().__init__(name, **kwargs)

        self._address = address
        self._port = port
        self._timeout = timeout
        self._terminator = terminator
        self._confirmation = write_confirmation

        self._ensure_connection = EnsureConnection(self)
        self._buffer_size = 1400

        self._socket: Optional[socket.socket] = None

        self.set_persistent(persistent)

    def set_address(self, address: Optional[str] = None,
                    port: Optional[int] = None) -> None:
        """
        Change the IP address and/or port of this instrument.

        Args:
            address: The IP address or name.
            port: The IP port.
        """
        if address is not None:
            self._address = address
        elif not hasattr(self, '_address'):
            raise TypeError('This instrument doesn\'t have an address yet, '
                            'you must provide one.')
        if port is not None:
            self._port = port
        elif not hasattr(self, '_port'):
            raise TypeError('This instrument doesn\'t have a port yet, '
                            'you must provide one.')

        self._disconnect()
        self.set_persistent(self._persistent)

    def set_persistent(self, persistent: bool) -> None:
        """
        Change whether this instrument keeps its socket open between calls.

        Args:
            persistent: Set True to keep the socket open all the time.
        """
        self._persistent = persistent
        if persistent:
            self._connect()
        else:
            self._disconnect()

    def flush_connection(self) -> None:
        self._recv()

    def _connect(self) -> None:
        if self._socket is not None:
            self._disconnect()

        try:
            log.info("Opening socket")
            self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            log.info("Connecting socket to {}:{}".format(self._address,
                                                         self._port))
            self._socket.connect((self._address, self._port))
            self.set_timeout(self._timeout)
        except ConnectionRefusedError:
            log.warning("Socket connection failed")
            if self._socket is not None:
                self._socket.close()
            self._socket = None
            raise

    def _disconnect(self) -> None:
        if self._socket is None:
            return
        log.info("Socket shutdown")
        self._socket.shutdown(socket.SHUT_RDWR)
        log.info("Socket closing")
        self._socket.close()
        log.info("Socket closed")
        self._socket = None

    def set_timeout(self, timeout: float) -> None:
        """
        Change the read timeout for the socket.

        Args:
            timeout: Seconds to allow for responses.
        """
        self._timeout = timeout

        if self._socket is not None:
            self._socket.settimeout(float(self._timeout))

    def set_terminator(self, terminator: str) -> None:
        r"""
        Change the write terminator to use.

        Args:
            terminator: Character(s) to terminate each send.
                Default '\n'.
        """
        self._terminator = terminator

    def _send(self, cmd: str) -> None:
        if self._socket is None:
            raise RuntimeError(f'IPInstrument {self.name} is not connected')
        data = cmd + self._terminator
        log.debug(f"Writing {data} to instrument {self.name}")
        self._socket.sendall(data.encode())

    def _recv(self) -> str:
        if self._socket is None:
            raise RuntimeError(f'IPInstrument {self.name} is not connected')
        result = self._socket.recv(self._buffer_size)
        log.debug(f"Got {result!r} from instrument {self.name}")
        if result == b'':
            log.warning("Got empty response from Socket recv() "
                        "Connection broken.")
        return result.decode()

    def close(self) -> None:
        """Disconnect and irreversibly tear down the instrument."""
        self._disconnect()
        super().close()

    def write_raw(self, cmd: str) -> None:
        """
        Low-level interface to send a command that gets no response.

        Args:
            cmd: The command to send to the instrument.
        """

        with self._ensure_connection:
            self._send(cmd)
            if self._confirmation:
                self._recv()

    def ask_raw(self, cmd: str) -> str:
        """
        Low-level interface to send a command an read a response.

        Args:
            cmd: The command to send to the instrument.

        Returns:
            The instrument's string response.
        """
        with self._ensure_connection:
            self._send(cmd)
            return self._recv()

    def snapshot_base(self, update: Optional[bool] = False,
                      params_to_skip_update: Optional[Sequence[str]] = None
                      ) -> Dict[Any, Any]:
        """
        State of the instrument as a JSON-compatible dict (everything that
        the custom JSON encoder class
        :class:`qcodes.utils.helpers.NumpyJSONEncoder`
        supports).

        Args:
            update: If True, update the state by querying the
                instrument. If None only update if the state is known to be
                invalid. If False, just use the latest values in memory and
                never update.
            params_to_skip_update: List of parameter names that will be
                skipped in update even if update is True. This is useful
                if you have parameters that are slow to update but can
                be updated in a different way (as in the qdac). If you
                want to skip the update of certain parameters in all
                snapshots, use the `snapshot_get`  attribute of those
                parameters: instead.

        Returns:
            dict: base snapshot
        """
        snap = super().snapshot_base(
            update=update,
            params_to_skip_update=params_to_skip_update)

        snap['port'] = self._port
        snap['confirmation'] = self._confirmation
        snap['address'] = self._address
        snap['terminator'] = self._terminator
        snap['timeout'] = self._timeout
        snap['persistent'] = self._persistent

        return snap


class EnsureConnection:

    """
    Context manager to ensure an instrument is connected when needed.

    Uses ``instrument._persistent`` to determine whether or not to close
    the connection immediately on completion.

    Args:
        instrument: the instance to connect.
    """

    def __init__(self, instrument: IPInstrument):
        self.instrument = instrument

    def __enter__(self) -> None:
        """Make sure we connect when entering the context."""
        if not self.instrument._persistent or self.instrument._socket is None:
            self.instrument._connect()

    def __exit__(self,
                 exc_type: Optional[Type[BaseException]],
                 exc_value: Optional[BaseException],
                 traceback: Optional[TracebackType]) -> None:
        """Possibly disconnect on exiting the context."""
        if not self.instrument._persistent:
            self.instrument._disconnect()



"""
Module containing all objects used for the description of a run.

Since run descriptions are persisted to disk, we have a strong obligation to
ensure backwards compatibility. The code is organised such that everything at
this module's top level is the CURRENT version. Older objects and the functions
necessary to convert between different versions are found in
:mod:`.versioning`.
"""



"""
This module contains helper functions that provide information about how
QCoDeS is installed and about what other packages are installed along with
QCoDeS
"""
import json
import logging
import subprocess
import sys
from typing import Dict, List, Optional

import pkg_resources
import requirements

if sys.version_info >= (3, 8):
    from importlib.metadata import PackageNotFoundError, distribution, version
else:
    # 3.7 and earlier
    from importlib_metadata import distribution, version, PackageNotFoundError

import qcodes

log = logging.getLogger(__name__)


def is_qcodes_installed_editably() -> Optional[bool]:
    """
    Try to ask pip whether QCoDeS is installed in editable mode and return
    the answer a boolean. Returns None if pip somehow did not respond as
    expected.
    """

    answer: Optional[bool]

    try:
        pipproc = subprocess.run(['python', '-m', 'pip', 'list', '-e', '--no-index',
                                  '--format=json'],
                                 check=True,
                                 stdout=subprocess.PIPE)
        e_pkgs = json.loads(pipproc.stdout.decode('utf-8'))
        answer = any([d["name"] == 'qcodes' for d in e_pkgs])
    except Exception as e:  # we actually do want a catch-all here
        log.warning(f'{type(e)}: {str(e)}')
        answer = None

    return answer


def get_qcodes_version() -> str:
    """
    Get the version of the currently installed QCoDeS
    """
    return qcodes.__version__


def get_qcodes_requirements() -> List[str]:
    """
    Return a list of the names of the packages that QCoDeS requires
    """
    qc_pkg = distribution('qcodes').requires
    if qc_pkg is None:
        return []
    package_names = [list(requirements.parse(req))[0].name for req in qc_pkg]

    return package_names


def get_qcodes_requirements_versions() -> Dict[str, str]:
    """
    Return a dictionary of the currently installed versions of the packages
    that QCoDeS requires. The dict maps package name to version string.
    If an (optional) dependency is not installed the name maps to "Not installed".
    """

    req_names = get_qcodes_requirements()

    req_versions = {}

    for req in req_names:
        try:
            req_versions[req] = version(req)
        except PackageNotFoundError:
            req_versions[req] = "Not installed"

    return req_versions


def get_all_installed_package_versions() -> Dict[str, str]:
    """
    Return a dictionary of the currently installed packages and their versions.
    """
    packages = pkg_resources.working_set
    return {i.key: i.version for i in packages}



# Test some subscription scenarios
from typing import List, Tuple, Dict, Union, Any
from numbers import Number

import pytest
from numpy import ndarray
import logging

import qcodes
from qcodes.dataset.descriptions.param_spec import ParamSpecBase
from qcodes.dataset.descriptions.dependencies import InterDependencies_
from qcodes.dataset.sqlite.connection import atomic_transaction

from qcodes.tests.common import default_config
from qcodes.tests.common import retry_until_does_not_throw


log = logging.getLogger(__name__)

VALUE = Union[str, Number, List[Any], ndarray, bool]


class MockSubscriber():
    """
    A basic subscriber factory that creates a subscriber, that
    just puts results and length into state.
    *Important*
    This class is extremely dangerous! Within the callback,
    you cannot read or write to the database/dataset because it
    is called from another thread than the one holding the connection of the
    dataset!
    """
    def __init__(self, ds, lg):
        self.lg = lg
        self.ds = ds

    def __call__(self, results: List[Tuple[VALUE]],
                 length: int, state: Dict[Any, Any]) -> None:
        log.debug(f'got log {self.lg} and dataset {self.ds.completed}.')
        state[length] = results


def config_subscriber_factory(ds, l):
    def config_subscriber(results, length, state):
        state[length] = results
        log.debug(f'got log {l} and dataset {ds.completed}.')
    return config_subscriber


@pytest.fixture(scope='function')
def basic_subscriber():
    """
    A basic subscriber that just puts results and length into
    state
    """

    def subscriber(results: List[Tuple[VALUE]], length: int,
                   state: Dict) -> None:
        state[length] = results

    return subscriber


@pytest.mark.flaky(reruns=5)
@pytest.mark.serial
def test_basic_subscription(dataset, basic_subscriber):
    xparam = ParamSpecBase(name='x',
                           paramtype='numeric',
                           label='x parameter',
                           unit='V')
    yparam = ParamSpecBase(name='y',
                           paramtype='numeric',
                           label='y parameter',
                           unit='Hz')
    idps = InterDependencies_(dependencies={yparam: (xparam,)})
    dataset.set_interdependencies(idps)
    dataset.mark_started()

    sub_id = dataset.subscribe(basic_subscriber, min_wait=0, min_count=1,
                               state={})

    assert len(dataset.subscribers) == 1
    assert list(dataset.subscribers.keys()) == [sub_id]

    expected_state = {}

    for x in range(10):
        y = -x**2
        dataset.add_results([{'x': x, 'y': y}])
        expected_state[x+1] = [(x, y)]

        @retry_until_does_not_throw(
            exception_class_to_expect=AssertionError, delay=0.5, tries=10)
        def assert_expected_state():
            assert dataset.subscribers[sub_id].state == expected_state

        assert_expected_state()

    dataset.unsubscribe(sub_id)

    assert len(dataset.subscribers) == 0
    assert list(dataset.subscribers.keys()) == []

    # Ensure the trigger for the subscriber has been removed from the database
    get_triggers_sql = "SELECT * FROM sqlite_master WHERE TYPE = 'trigger';"
    triggers = atomic_transaction(
        dataset.conn, get_triggers_sql).fetchall()
    assert len(triggers) == 0


def test_subscription_from_config(dataset, basic_subscriber):
    """
    This test is similar to `test_basic_subscription`, with the only
    difference that another subscriber from a config file is added.
    """
    # This string represents the config file in the home directory:
    config = """
    {
        "subscription":{
            "subscribers":{
                "test_subscriber":{
                    "factory": "qcodes.tests.dataset.test_subscribing.MockSubscriber",
                    "factory_kwargs":{
                        "lg": false
                    },
                    "subscription_kwargs":{
                        "min_wait": 0,
                        "min_count": 1,
                        "callback_kwargs": {}
                    }
                }
            }
        }
    }
    """
    # This little dance around the db_location is due to the fact that the
    # dataset fixture creates a dataset in a db in a temporary directory.
    # Therefore we need to 'backup' the path to the db when using the
    # default configuration.
    db_location = qcodes.config.core.db_location
    with default_config(user_config=config):
        qcodes.config.core.db_location = db_location

        assert 'test_subscriber' in qcodes.config.subscription.subscribers

        xparam = ParamSpecBase(name='x',
                           paramtype='numeric',
                           label='x parameter',
                           unit='V')
        yparam = ParamSpecBase(name='y',
                              paramtype='numeric',
                              label='y parameter',
                              unit='Hz')
        idps = InterDependencies_(dependencies={yparam: (xparam,)})
        dataset.set_interdependencies(idps)

        dataset.mark_started()

        sub_id = dataset.subscribe(basic_subscriber, min_wait=0, min_count=1,
                                   state={})
        sub_id_c = dataset.subscribe_from_config('test_subscriber')
        assert len(dataset.subscribers) == 2
        assert list(dataset.subscribers.keys()) == [sub_id, sub_id_c]

        expected_state = {}

        # Here we are only testing 2 to reduce the CI time
        for x in range(2):
            y = -x**2
            dataset.add_results([{'x': x, 'y': y}])
            expected_state[x+1] = [(x, y)]

            @retry_until_does_not_throw(
                exception_class_to_expect=AssertionError, tries=10)
            def assert_expected_state():
                assert dataset.subscribers[sub_id].state == expected_state
                assert dataset.subscribers[sub_id_c].state == expected_state

            assert_expected_state()


def test_subscription_from_config_wrong_name(dataset):
    """
    This test checks that an exception is thrown if a wrong name for a
    subscriber is passed
    """
    # This string represents the config file in the home directory:
    config = """
    {
        "subscription":{
            "subscribers":{
                "test_subscriber_wrong":{
                    "factory": "qcodes.tests.dataset.test_subscribing.MockSubscriber",
                    "factory_kwargs":{
                        "lg": false
                    },
                    "subscription_kwargs":{
                        "min_wait": 0,
                        "min_count": 1,
                        "callback_kwargs": {}
                    }
                }
            }
        }
    }
    """
    db_location = qcodes.config.core.db_location
    with default_config(user_config=config):
        qcodes.config.core.db_location = db_location

        assert 'test_subscriber' not in qcodes.config.subscription.subscribers
        with pytest.raises(RuntimeError):
            sub_id_c = dataset.subscribe_from_config('test_subscriber')



from typing import Any

import pytest

from qcodes import Instrument, InstrumentChannel
from qcodes.instrument.base import InstrumentBase


class ExampleBaseVoltageSource(Instrument):
    """
    All abstract parameters *must* be implemented
    before this class can be initialized. This
    allows us to enforce an interface.
    """

    def __init__(self, name: str):
        super().__init__(name)

        self.add_parameter(
            "voltage", unit="V", abstract=True, get_cmd=None, set_cmd=None
        )

        self.add_parameter("current", unit="A", get_cmd=None, set_cmd=None)


class VoltageSource(ExampleBaseVoltageSource):
    """
    Make a specific implementation of the interface
    """

    def __init__(self, name: str):
        super().__init__(name)

        self.add_parameter("voltage", unit="V", get_cmd=None, set_cmd=None)


class VoltageSourceNotImplemented(ExampleBaseVoltageSource):
    """
    We 'forget' to implement the voltage parameter
    """


class VoltageSourceBadUnit(ExampleBaseVoltageSource):
    """
    The units must match between sub and base classes
    """

    def __init__(self, name: str):
        super().__init__(name)

        self.add_parameter(
            "voltage", unit="mV", get_cmd=None, set_cmd=None  # This should be 'V'
        )


class VoltageSourceInitException(Instrument):
    """
    We conditionally raise an assertion error in the init.
    The instrument should not be registered when one
    is raised
    """

    def __init__(self, name: str, do_raise=True):
        super().__init__(name)

        if do_raise:
            assert False


class VoltageSourceSubSub(VoltageSource):
    """
    This is a sub-sub class of the example base voltage
    source. The post init function should be called
    only once
    """

    call_count = 0

    def __init__(self, name: str):
        super().__init__(name)

    def __post_init__(self, *args: Any, **kwargs: Any) -> None:
        super().__post_init__()  # type: ignore[misc]
        self.call_count += 1


class VoltageChannelBase(InstrumentChannel):
    """
    Create a channel with an abstract parameter
    """

    def __init__(self, parent: InstrumentBase, name: str, **kwargs: Any):
        super().__init__(parent, name, **kwargs)

        self.add_parameter("voltage", unit="V", abstract=True)


class VoltageChannel(VoltageChannelBase):
    """
    Create an implementation of the abstract channel
    """

    def __init__(self, parent: InstrumentBase, name: str, **kwargs: Any):
        super().__init__(parent, name, **kwargs)

        self.add_parameter("voltage", unit="V", get_cmd=None, set_cmd=None)


@pytest.fixture(name="driver", scope="module")
def _driver():
    drvr = VoltageSource("driver")
    yield drvr
    drvr.close()


def test_sanity(driver):
    """
    If all abstract parameters are implemented, we should be able
    to instantiate the instrument
    """
    driver.voltage(0.1)
    assert driver.voltage() == 0.1


@pytest.mark.skip("tests unimplemented feature")
def test_not_implemented_error():
    """
    If not all abstract parameters are implemented, we should see
    an exception
    """
    with pytest.raises(
        NotImplementedError, match="has un-implemented Abstract Parameter"
    ):
        VoltageSourceNotImplemented("driver2")
    assert not VoltageSourceNotImplemented.instances()


def test_get_set_raises():
    """
    If not all abstract parameters are implemented, we should see
    an exception
    """
    vs = VoltageSourceNotImplemented("driver2")

    with pytest.raises(
        NotImplementedError, match="Trying to get an abstract parameter"
    ):
        vs.voltage.get()

    with pytest.raises(
        NotImplementedError, match="Trying to set an abstract parameter"
    ):
        vs.voltage.set(0)


def test_unit_value_error():
    """
    Units should match between subclasses and base classes
    """
    with pytest.raises(ValueError, match="This is inconsistent with the unit defined"):
        VoltageSourceBadUnit("driver3")


@pytest.mark.xfail()
def test_unit_value_error_does_not_register_instrument():
    """
    Units should match between subclasses and base classes
    """
    with pytest.raises(ValueError, match="This is inconsistent with the unit defined"):
        VoltageSourceBadUnit("driver3a")
    assert not VoltageSourceBadUnit.instances()


@pytest.mark.skip("tests unimplemented feature")
def test_exception_in_init():
    """
    In previous versions of QCoDeS, if an error occurred in
    the instrument init method, we could not attempt to retry
    the instantiation with the same instrument name. This is
    because the driver instance was recorded at the end of
    the init method in the base class. In current versions
    the instance is recorded in the __post_init__ method,
    which should eliminate this problem
    """
    name = "driver4"
    try:
        VoltageSourceInitException(name)
    except AssertionError:
        pass
    assert not VoltageSourceInitException.instances()
    instance = VoltageSourceInitException(name, do_raise=False)
    assert name in [ins.name for ins in VoltageSourceInitException.instances()]
    instance.close()


@pytest.mark.skip("tests unimplemented feature")
def test_subsub():
    """
    Verify that the post init method is only called once, even
    for sub-sub classes. This should work for arbitrary levels
    of subclassing.
    """
    instance = VoltageSourceSubSub("driver5")
    assert instance.call_count == 1


@pytest.mark.skip("tests unimplemented feature")
def test_channel(driver):
    """
    This should work without exceptions
    """
    VoltageChannel(driver, "driver6")

    with pytest.raises(
        NotImplementedError, match="has un-implemented Abstract Parameter"
    ):
        VoltageChannelBase(driver, "driver6")



import pytest
from contextlib import contextmanager
import tempfile
import warnings
from pathlib import Path
import os
from typing import Optional
import json
from io import StringIO

import qcodes
import qcodes.utils.validators as validators
from qcodes.utils.helpers import get_qcodes_path
from qcodes.utils.deprecate import assert_deprecated, deprecation_message
from qcodes.instrument.parameter import DelegateParameter
from qcodes import Instrument
from qcodes.station import (
    Station, ValidationWarning, update_config_schema, SCHEMA_PATH)
from qcodes.instrument.parameter import Parameter
from qcodes.monitor.monitor import Monitor
from qcodes.tests.instrument_mocks import (
    DummyInstrument)
from qcodes.tests.common import default_config
from qcodes.utils.helpers import NumpyJSONEncoder
from qcodes.utils.helpers import YAML
from .common import DumyPar


@pytest.fixture(autouse=True)
def use_default_config():
    with default_config():
        yield


@pytest.fixture(autouse=True)
def set_default_station_to_none():
    """Makes sure that after startup and teardown there is no default station"""
    Station.default = None
    yield
    Station.default = None


@pytest.fixture(autouse=True)
def close_all_instruments():
    """Makes sure that after startup and teardown all instruments are closed"""
    Instrument.close_all()
    yield
    Instrument.close_all()


@pytest.fixture(autouse=True)
def treat_validation_warning_as_error():
    warnings.simplefilter("error", ValidationWarning)
    yield
    warnings.simplefilter("default", ValidationWarning)


def test_station():
    bob = DummyInstrument('bob', gates=['one'])
    station = Station(bob)

    assert ['bob'] == list(station.components.keys())
    assert bob == station.components['bob']

    assert station == station.default
    assert station == Station.default


def test_station_getitem():
    bob = DummyInstrument('bob', gates=['one'])
    station = Station(bob)

    assert bob == station['bob']

    with pytest.raises(KeyError, match='bobby'):
        _ = station.components['bobby']


def test_station_delegated_attributes():
    bob = DummyInstrument('bob', gates=['one'])
    station = Station(bob)

    assert bob == station.bob

    with pytest.raises(AttributeError, match="'Station' object and its "
                                             "delegates have no attribute "
                                             "'bobby'"):
        _ = station.bobby


def test_add_component():
    bob = DummyInstrument('bob', gates=['one'])
    station = Station()
    station.add_component(bob, 'bob')

    assert ['bob'] == list(station.components.keys())
    assert bob == station.components['bob']


def test_add_component_without_specifying_name():
    """
    Test that station looks for 'name' attribute in the component and uses it
    """
    bob = DummyInstrument('bob', gates=['one'])
    assert hasattr(bob, 'name')
    assert 'bob' == bob.name

    station = Station()
    station.add_component(bob)

    assert ['bob'] == list(station.components.keys())
    assert bob == station.components['bob']


def test_add_component_with_no_name():
    """
    Test that station comes up with a name for components without 'name'
    attribute
    """
    bob = {'name', 'bob'}
    station = Station()
    station.add_component(bob)

    assert ['component0'] == list(station.components.keys())
    assert bob == station.components['component0']

    jay = {'name', 'jay'}
    station.add_component(jay)

    assert ['component0', 'component1'] == list(station.components.keys())
    assert jay == station.components['component1']


def test_remove_component():
    bob = DummyInstrument('bob', gates=['one'])
    station = Station()
    station.add_component(bob, 'bob')

    assert ['bob'] == list(station.components.keys())
    assert bob == station.components['bob']

    bob2 = station.remove_component('bob')

    with pytest.raises(KeyError, match='bob'):
        _ = station.components['bob']
    assert bob == bob2

    with pytest.raises(KeyError, match='Component bobby is not part of the '
                                       'station'):
        _ = station.remove_component('bobby')


def test_close_all_registered_instruments():
    names = [f'some_name_{i}' for i in range(10)]
    instrs = [Instrument(name=name) for name in names]
    st = Station(*instrs)
    for name in names:
        assert name in Instrument._all_instruments
    st.close_all_registered_instruments()
    for name in names:
        assert name not in Instrument._all_instruments


def test_snapshot():
    station = Station()

    empty_snapshot = station.snapshot()
    assert {'instruments': {},
            'parameters': {},
            'components': {},
            'config': None,
            } == empty_snapshot

    instrument = DummyInstrument('instrument', gates=['one'])
    station.add_component(instrument)
    instrument_snapshot = instrument.snapshot()

    parameter = Parameter('parameter', label='Label', unit='m')
    station.add_component(parameter)
    parameter_snapshot = parameter.snapshot()

    excluded_parameter = Parameter('excluded_parameter', snapshot_exclude=True)
    station.add_component(excluded_parameter)

    component = DumyPar('component')
    component.metadata['smth'] = 'in the way she moves'
    station.add_component(component)
    component_snapshot = component.snapshot()

    snapshot = station.snapshot()

    assert isinstance(snapshot, dict)
    assert ['instruments',
            'parameters',
            'components',
            'config',
            ] == list(snapshot.keys())

    assert ['instrument'] == list(snapshot['instruments'].keys())
    assert instrument_snapshot == snapshot['instruments']['instrument']

    # the list should not contain the excluded parameter
    assert ['parameter'] == list(snapshot['parameters'].keys())
    assert parameter_snapshot == snapshot['parameters']['parameter']

    assert ['component'] == list(snapshot['components'].keys())
    assert component_snapshot == snapshot['components']['component']


def test_station_after_instrument_is_closed():
    """
    Test that station is aware of the fact that its components could be
    removed within the lifetime of the station. Here we instantiate an
    instrument, add it to a station, then close the instrument, and then
    perform an action on the station to ensure that the closed instrument
    does not break the work of the station object.
    """
    bob = DummyInstrument('bob', gates=['one'])

    station = Station(bob)

    assert bob == station['bob']

    bob.close()

    # 'bob' is closed, but it is still part of the station
    assert bob == station['bob']

    # check that snapshot method executes without exceptions
    snapshot = station.snapshot()

    # check that 'bob's snapshot is not here (because 'bob' is closed,
    # hence it was ignored, and even removed from the station by
    # `snapshot_base` method)
    assert {'instruments': {},
            'parameters': {},
            'components': {},
            'config': None,
            } == snapshot

    # check that 'bob' has been removed from the station
    with pytest.raises(KeyError, match='bob'):
        _ = station.components['bob']

    # check that 'bob' has been removed from the station, again
    with pytest.raises(KeyError, match='Component bob is not part of the '
                                       'station'):
        station.remove_component('bob')


def test_update_config_schema():
    update_config_schema()
    with open(SCHEMA_PATH) as f:
        schema = json.load(f)
    assert len(schema['definitions']['instruments']['enum']) > 1


@contextmanager
def config_file_context(file_content):
    with tempfile.TemporaryDirectory() as tmpdirname:
        filename = Path(tmpdirname, 'station_config.yaml')
        with filename.open('w') as f:
            f.write(file_content)
        yield str(filename)


@contextmanager
def config_files_context(file_content1, file_content2):
    with tempfile.TemporaryDirectory() as tmpdirname:
        filename1 = Path(tmpdirname, 'station_config1.yaml')
        with filename1.open('w') as f:
            f.write(file_content1)
        filename2 = Path(tmpdirname, 'station_config2.yaml')
        with filename2.open('w') as f:
            f.write(file_content2)
        yield [str(filename1), str(filename2)]


@pytest.fixture(name="example_station_config")
def _make_example_station_config():
    """
    Returns path to temp yaml file with station config.
    """
    sims_path = get_qcodes_path('instrument', 'sims')
    test_config = f"""
instruments:
  lakeshore:
    type: qcodes.instrument_drivers.Lakeshore.Model_336.Model_336
    enable_forced_reconnect: true
    address: GPIB::2::INSTR
    init:
      visalib: '{sims_path}lakeshore_model336.yaml@sim'
  mock_dac:
    type: qcodes.tests.instrument_mocks.DummyInstrument
    enable_forced_reconnect: true
    init:
      gates: {{"ch1", "ch2"}}
    parameters:
      ch1:
        monitor: true
  mock_dac2:
    type: qcodes.tests.instrument_mocks.DummyInstrument
    """
    with config_file_context(test_config) as filename:
        yield filename


def test_dynamic_reload_of_file(example_station_config):
    st = Station(config_file=example_station_config)
    mock_dac = st.load_instrument('mock_dac')
    assert 'ch1' in mock_dac.parameters
    with open(example_station_config) as f:
        filedata = f.read().replace('ch1', 'gate1')
    with open(example_station_config, 'w') as f:
        f.write(filedata)
    mock_dac = st.load_instrument('mock_dac')
    assert 'ch1' not in mock_dac.parameters
    assert 'gate1' in mock_dac.parameters


def station_from_config_str(config: str) -> Station:
    st = Station(config_file=None)
    st.load_config(config)
    return st


def station_config_has_been_loaded(st: Station) -> bool:
    return st.config is not None


@pytest.fixture(name="example_station")
def _make_example_station(example_station_config):
    return Station(config_file=example_station_config)


# instrument loading related tests
def test_station_config_path_resolution(example_station_config):
    config = qcodes.config["station"]

    # There is no default yaml file present that defines a station
    # so we expect the station config not to be loaded.
    assert not station_config_has_been_loaded(Station())

    path = Path(example_station_config)
    config["default_file"] = str(path)
    # Now the default file with the station configuration is specified, and
    # this file exists, hence we expect the Station to have the station
    # configuration loaded upon initialization.
    assert station_config_has_been_loaded(Station())

    config["default_file"] = path.name
    config["default_folder"] = str(path.parent)
    # Here the default_file setting contains only the file name, and the
    # default_folder contains the path to the folder where this file is
    # located, hence we again expect that the station configuration is loaded
    # upon station initialization.
    assert station_config_has_been_loaded(Station())

    config["default_file"] = 'random.yml'
    config["default_folder"] = str(path.parent)
    # In this case, the station configuration file specified in the qcodes
    # config does not exist, hence the initialized station is not expected to
    # have station configuration loaded.
    assert not station_config_has_been_loaded(Station())

    config["default_file"] = str(path)
    config["default_folder"] = r'C:\SomeOtherFolder'
    # In this case, the default_file setting of the qcodes config contains
    # absolute path to the station configuration file, while the default_folder
    # setting is set to some non-existent folder.
    # In this situation, the value of the default_folder will be ignored,
    # but because the file specified in default_file setting exists,
    # the station will be initialized with the loaded configuration.
    assert station_config_has_been_loaded(Station())

    config["default_file"] = None
    config["default_folder"] = str(path.parent)
    # When qcodes config has only the default_folder setting specified to an
    # existing folder, and default_file setting is not specified, then
    # passing the name of a station configuration file, that exists in that
    # default_folder, as an argument to the Station is expected to result
    # in a station with loaded configuration.
    assert station_config_has_been_loaded(Station(config_file=path.name))

    config["default_file"] = None
    config["default_folder"] = None
    # In case qcodes config does not have default_file and default_folder
    # settings specified, passing an absolute file path as an argument to the
    # station is expected to result in a station with loaded configuration.
    assert station_config_has_been_loaded(Station(config_file=str(path)))


def test_station_configuration_is_a_component_of_station(example_station):
    assert station_config_has_been_loaded(example_station)


def test_station_config_can_be_loaded_from_snapshot(example_station):
    assert station_config_has_been_loaded(example_station)
    # ensure that we can correctly dump config which is a subclass of UserDict
    configdump = json.dumps(example_station.config, cls=NumpyJSONEncoder)
    # as this is now a regular dict we can load it back
    loaded_config = json.loads(configdump)
    # now lets ensure that we can recreate the
    # station from the loaded config
    # first we need to get a yaml repr of the data
    yaml = YAML()
    with StringIO() as output:
        yaml.dump(loaded_config, output)
        yaml_repr = output.getvalue()
    # which we can then reload into the station
    new_station = Station(default=False)
    new_station.load_config(yaml_repr)
    assert example_station.config == new_station.config


@pytest.fixture
def simple_mock_station():
    yield station_from_config_str(
        """
instruments:
  mock:
    type: qcodes.tests.instrument_mocks.DummyInstrument
        """)


def test_simple_mock_config(simple_mock_station):
    st = simple_mock_station
    assert station_config_has_been_loaded(st)
    assert hasattr(st, 'load_mock')
    mock_snapshot = st.snapshot()['config']['instruments']['mock']
    assert (mock_snapshot['type'] ==
            "qcodes.tests.instrument_mocks.DummyInstrument")
    assert 'mock' in st.config['instruments']


def test_simple_mock_load_mock(simple_mock_station):
    st = simple_mock_station
    mock = st.load_mock()
    assert isinstance(mock, DummyInstrument)
    assert mock.name == 'mock'
    assert st.components['mock'] is mock


def test_simple_mock_load_instrument(simple_mock_station):
    st = simple_mock_station
    mock = st.load_instrument('mock')
    assert isinstance(mock, DummyInstrument)
    assert mock.name == 'mock'
    assert st.components['mock'] is mock


def test_enable_force_reconnect() -> None:
    def get_instrument_config(enable_forced_reconnect: Optional[bool]) -> str:
        return f"""
instruments:
  mock:
    type: qcodes.tests.instrument_mocks.DummyInstrument
    {f'enable_forced_reconnect: {enable_forced_reconnect}'
        if enable_forced_reconnect is not None else ''}
    init:
      gates: {{"ch1", "ch2"}}
         """

    def assert_on_reconnect(*, use_user_cfg: Optional[bool],
                            use_instr_cfg: Optional[bool],
                            expect_failure: bool) -> None:
        qcodes.config["station"]\
            ['enable_forced_reconnect'] = use_user_cfg
        st = station_from_config_str(
            get_instrument_config(use_instr_cfg))
        st.load_instrument('mock')
        if expect_failure:
            with pytest.raises(KeyError) as excinfo:
                st.load_instrument('mock')
            assert ("Another instrument has the name: mock"
                    in str(excinfo.value))
        else:
            st.load_instrument('mock')
        Instrument.close_all()

    for use_user_cfg in [None, True, False]:
        assert_on_reconnect(use_user_cfg=use_user_cfg,
                            use_instr_cfg=False,
                            expect_failure=True)
        assert_on_reconnect(use_user_cfg=use_user_cfg,
                            use_instr_cfg=True,
                            expect_failure=False)

    assert_on_reconnect(use_user_cfg=True,
                        use_instr_cfg=None,
                        expect_failure=False)
    assert_on_reconnect(use_user_cfg=False,
                        use_instr_cfg=None,
                        expect_failure=True)
    assert_on_reconnect(use_user_cfg=None,
                        use_instr_cfg=None,
                        expect_failure=True)


def test_revive_instance():
    st = station_from_config_str("""
instruments:
  mock:
    type: qcodes.tests.instrument_mocks.DummyInstrument
    enable_forced_reconnect: true
    init:
      gates: {"ch1"}
    """)
    mock = st.load_instrument('mock')
    mock2 = st.load_instrument('mock')
    assert mock is not mock2
    assert mock is not st.mock
    assert mock2 is st.mock

    mock3 = st.load_instrument('mock', revive_instance=True)
    assert mock3 == mock2
    assert mock3 == st.mock


def test_init_parameters():
    st = station_from_config_str(
        """
instruments:
  mock:
    type: qcodes.tests.instrument_mocks.DummyInstrument
    enable_forced_reconnect: true
    init:
      gates: {"ch1", "ch2"}
    """
    )
    mock = st.load_instrument('mock')
    for ch in ["ch1", "ch2"]:
        assert ch in mock.parameters.keys()
    assert len(mock.parameters) == 3  # there is also IDN

    # Overwrite parameter
    mock = st.load_instrument('mock', gates=["TestGate"])
    assert "TestGate" in mock.parameters.keys()
    assert len(mock.parameters) == 2  # there is also IDN

    # test address
    sims_path = get_qcodes_path('instrument', 'sims')
    st = station_from_config_str(f"""
instruments:
  lakeshore:
    type: qcodes.instrument_drivers.Lakeshore.Model_336.Model_336
    enable_forced_reconnect: true
    address: GPIB::2::INSTR
    init:
      visalib: '{sims_path}lakeshore_model336.yaml@sim'
    """)
    st.load_instrument('lakeshore')


def test_name_init_kwarg(simple_mock_station):
    # special case of `name` as kwarg
    st = simple_mock_station
    mock = st.load_instrument('mock', name='test')
    assert mock.name == 'test'
    assert st.components['test'] is mock


def test_name_specified_in_init_in_yaml_is_used():
    st = station_from_config_str(
        """
instruments:
  mock:
    type: qcodes.tests.instrument_mocks.DummyInstrument
    init:
      name: dummy
        """)

    mock = st.load_instrument('mock')
    assert isinstance(mock, DummyInstrument)
    assert mock.name == 'dummy'
    assert st.components['dummy'] is mock


class InstrumentWithNameAsNotFirstArgument(Instrument):
    def __init__(self, first_arg, name):
        super().__init__(name)
        self._first_arg = first_arg


def test_able_to_load_instrument_with_name_argument_not_being_the_first():
    st = station_from_config_str(
        """
instruments:
  name_goes_second:
    type: qcodes.tests.test_station.InstrumentWithNameAsNotFirstArgument
        """)

    instr = st.load_instrument('name_goes_second', first_arg=42)
    assert isinstance(instr, InstrumentWithNameAsNotFirstArgument)
    assert instr.name == 'name_goes_second'
    assert st.components['name_goes_second'] is instr


def test_setup_alias_parameters():
    st = station_from_config_str("""
instruments:
  mock:
    type: qcodes.tests.instrument_mocks.DummyInstrument
    enable_forced_reconnect: true
    init:
      gates: {"ch1"}
    parameters:
      ch1:
        unit: mV
        label: main gate
        scale: 2
        offset: 1
        limits: [-10, 10]
        alias: gate_a
        initial_value: 9

    """)
    mock = st.load_instrument('mock')
    p = getattr(mock, 'gate_a')
    assert isinstance(p, qcodes.Parameter)
    assert p.unit == 'mV'
    assert p.label == 'main gate'
    assert p.scale == 2
    assert p.offset == 1
    assert isinstance(p.vals, validators.Numbers)
    assert str(p.vals) == '<Numbers -10<=v<=10>'
    assert p() == 9
    mock.ch1(1)
    assert p() == 1
    p(3)
    assert mock.ch1() == 3
    assert p.raw_value == 7
    assert mock.ch1.raw_value == 7


def test_setup_delegate_parameters():
    st = station_from_config_str("""
instruments:
  mock:
    type: qcodes.tests.instrument_mocks.DummyInstrument
    enable_forced_reconnect: true
    init:
      gates: {"ch1"}
    parameters:
      ch1:
        unit: V
        label: ch1
        scale: 1
        offset: 0
        limits: [-10, 10]
    add_parameters:
      gate_a:
        source: ch1
        unit: mV
        label: main gate
        scale: 2
        offset: 1
        limits: [-6.0 , 6.]
        initial_value: 2

    """)
    mock = st.load_instrument('mock')
    p = getattr(mock, 'gate_a')
    assert isinstance(p, DelegateParameter)
    assert p.unit == 'mV'
    assert p.label == 'main gate'
    assert p.scale == 2
    assert p.offset == 1
    assert isinstance(p.vals, validators.Numbers)
    assert str(p.vals) == '<Numbers -6.0<=v<=6.0>'
    assert p() == 2
    assert mock.ch1.unit == 'V'
    assert mock.ch1.label == 'ch1'
    assert mock.ch1.scale == 1
    assert mock.ch1.offset == 0
    assert isinstance(p.vals, validators.Numbers)
    assert str(mock.ch1.vals) == '<Numbers -10<=v<=10>'
    assert mock.ch1() == 5
    mock.ch1(7)
    assert p() == 3
    assert p.raw_value == 7
    assert mock.ch1.raw_value == 7
    assert (json.dumps(mock.ch1.snapshot()) ==
            json.dumps(p.snapshot()['source_parameter']))


def test_channel_instrument():
    """Test that parameters from instrument's submodule also get configured correctly"""
    st = station_from_config_str("""
instruments:
  mock:
    type: qcodes.tests.instrument_mocks.DummyChannelInstrument
    enable_forced_reconnect: true
    parameters:
      A.temperature:
        unit: mK
    add_parameters:
      T:
        source: A.temperature
      A.voltage:
        source: A.temperature
    """)
    mock = st.load_instrument('mock')
    assert mock.A.temperature.unit == 'mK'
    assert mock.T.unit == 'mK'
    assert mock.A.voltage.source is mock.A.temperature


def test_setting_channel_parameter():
    st = station_from_config_str("""
instruments:
  mock:
    type: qcodes.tests.instrument_mocks.DummyChannelInstrument
    parameters:
      channels.temperature:
          initial_value: 10
    """)
    mock = st.load_instrument('mock')
    assert mock.channels.temperature() == (10,) * 6


def test_monitor_not_loaded_by_default(example_station_config):
    st = Station(config_file=example_station_config)
    st.load_instrument('mock_dac')
    assert Monitor.running is None


def test_monitor_loaded_if_specified(example_station_config):
    st = Station(config_file=example_station_config, use_monitor=True)
    st.load_instrument('mock_dac')
    assert Monitor.running is not None
    assert len(Monitor.running._parameters) == 1
    assert Monitor.running._parameters[0].name == 'ch1'
    Monitor.running.stop()


def test_monitor_loaded_by_default_if_in_config(example_station_config):
    qcodes.config["station"]['use_monitor'] = True
    st = Station(config_file=example_station_config)
    st.load_instrument('mock_dac')
    assert Monitor.running is not None
    assert len(Monitor.running._parameters) == 1
    assert Monitor.running._parameters[0].name == 'ch1'
    Monitor.running.stop()


def test_monitor_not_loaded_if_specified(example_station_config):
    st = Station(config_file=example_station_config, use_monitor=False)
    st.load_instrument('mock_dac')
    assert Monitor.running is None


def test_deprecated_driver_keyword():
    st = station_from_config_str("""
instruments:
  mock:
    driver: qcodes.tests.instrument_mocks
    type: DummyChannelInstrument
    """)
    with assert_deprecated(
        deprecation_message(
            'use of the "driver"-keyword in the station configuration file',
            alternative='the "type"-keyword instead, prepending the driver value'
                        ' to it')):
        st.load_instrument('mock')


def test_deprecated_limits_keyword_as_string():
    st = station_from_config_str("""
instruments:
  mock:
    type: qcodes.tests.instrument_mocks.DummyInstrument
    init:
      gates: {"ch1"}
    parameters:
      ch1:
        limits: -10, 10
    """)
    with assert_deprecated(
        deprecation_message(
            'use of a comma separated string for the limits keyword',
            alternative='an array like "[lower_lim, upper_lim]"')
    ):
        st.load_instrument('mock')


def test_config_validation_failure():
    with pytest.raises(ValidationWarning):
        station_from_config_str("""
instruments:
  mock:
    driver: qcodes.tests.instrument_mocks.DummyInstrument
invalid_keyword:
  more_errors: 42
        """)


def test_config_validation_failure_on_file():
    with pytest.raises(ValidationWarning):
        test_config = """
instruments:
  mock:
    driver: qcodes.tests.instrument_mocks.DummyInstrument
invalid_keyword:
  more_errors: 42
    """
        with config_file_context(test_config) as filename:
            Station(config_file=filename)


def test_config_validation_comprehensive_config():
    Station(config_file=os.path.join(
        get_qcodes_path(), 'dist', 'tests', 'station', 'example.station.yaml')
    )


def test_load_all_instruments_raises_on_both_only_names_and_only_types_passed(
        example_station
):
    with pytest.raises(
        ValueError,
        match="It is an error to supply both ``only_names`` "
              "and ``only_types`` arguments.",
    ):
        example_station.load_all_instruments(only_names=(), only_types=())


def test_load_all_instruments_no_args(example_station):
    all_instruments_in_config = {"lakeshore", "mock_dac", "mock_dac2"}

    loaded_instruments = example_station.load_all_instruments()

    assert set(loaded_instruments) == all_instruments_in_config

    for instrument in all_instruments_in_config:
        assert instrument in example_station.components
        assert Instrument.exist(instrument)


def test_load_all_instruments_only_types(example_station):
    all_dummy_instruments = {"mock_dac", "mock_dac2"}

    loaded_instruments = example_station.load_all_instruments(
        only_types=("DummyInstrument",)
    )

    assert set(loaded_instruments) == all_dummy_instruments

    for instrument in all_dummy_instruments:
        assert instrument in example_station.components
        assert Instrument.exist(instrument)

    other_instruments = (
            set(example_station.config["instruments"].keys())
            - all_dummy_instruments
    )

    for instrument in other_instruments:
        assert instrument not in example_station.components
        assert not Instrument.exist(instrument)


def test_load_all_instruments_only_names(example_station):
    instruments_to_load = {"lakeshore", "mock_dac"}

    loaded_instruments = example_station.load_all_instruments(
        only_names=instruments_to_load
    )

    assert set(loaded_instruments) == instruments_to_load

    for instrument in loaded_instruments:
        assert instrument in example_station.components
        assert Instrument.exist(instrument)

    other_instruments = (
            set(example_station.config["instruments"].keys())
            - instruments_to_load
    )

    for instrument in other_instruments:
        assert instrument not in example_station.components
        assert not Instrument.exist(instrument)


def test_load_all_instruments_without_config_raises():
    station = Station()
    with pytest.raises(ValueError, match="Station has no config"):
        station.load_all_instruments()


def test_station_config_created_with_multiple_config_files():

    test_config1 = f"""
        instruments:
          mock_dac1:
            type: qcodes.tests.instrument_mocks.DummyInstrument
            enable_forced_reconnect: true
            init:
              gates: {{"ch1", "ch2"}}
            parameters:
              ch1:
                monitor: true
    """
    test_config2 = f"""
        instruments:
          mock_dac2:
            type: qcodes.tests.instrument_mocks.DummyInstrument
    """
    with config_files_context(
            test_config1, test_config2
    ) as file_list:
        assert station_config_has_been_loaded(Station(config_file=file_list))



import re
from datetime import datetime
from typing import Optional

import pytest

from qcodes import Instrument
from qcodes.instrument.group_parameter import Group, GroupParameter


@pytest.fixture(autouse=True)
def close_all_instruments():
    """Makes sure that after startup and teardown all instruments are closed"""
    Instrument.close_all()
    yield
    Instrument.close_all()


class Dummy(Instrument):
    def __init__(self, name: str,
                 initial_a: Optional[int] = None,
                 initial_b: Optional[int] = None,
                 scale_a: Optional[float] = None,
                 get_cmd: Optional[str] = "CMD?") -> None:
        super().__init__(name)

        self._a = 0
        self._b = 0
        self._get_cmd = get_cmd

        self.add_parameter(
            "a",
            get_parser=int,
            parameter_class=GroupParameter,
            docstring="Some succinct description",
            label="label",
            unit="SI",
            initial_value=initial_a,
            scale=scale_a
        )

        self.add_parameter(
            "b",
            get_parser=int,
            parameter_class=GroupParameter,
            docstring="Some succinct description",
            label="label",
            unit="SI",
            initial_value=initial_b
        )

        self.group = Group(
            [self.a, self.b],
            set_cmd="CMD {a}, {b}",
            get_cmd=get_cmd
        )

    def write(self, cmd: str) -> None:
        result = re.search("CMD (.*), (.*)", cmd)
        assert result is not None
        self._a, self._b = (int(i) for i in result.groups())

    def ask(self, cmd: str) -> str:
        assert cmd == self._get_cmd
        return ",".join(str(i) for i in [self._a, self._b])


def test_sanity():
    """
    Test that we can individually address parameters "a" and "b", which belong
    to the same group.
    """
    dummy = Dummy("dummy")

    assert dummy.a() == 0
    assert dummy.b() == 0

    dummy.a(3)
    dummy.b(6)

    assert dummy.a() == 3
    assert dummy.b() == 6

    dummy.b(10)
    assert dummy.a() == 3
    assert dummy.b() == 10


def test_raise_on_get_set_cmd():

    for arg in ["set_cmd", "get_cmd"]:
        kwarg = {arg: ""}

        with pytest.raises(ValueError) as e:
            GroupParameter(name="a", **kwarg)

        assert str(e.value) == "A GroupParameter does not use 'set_cmd' or " \
                               "'get_cmd' kwarg"


def test_raises_on_get_set_without_group():
    param = GroupParameter(name='b')

    with pytest.raises(RuntimeError) as e:
        param.get()
    assert str(e.value) == "('Trying to get Group value but no group defined', 'getting b')"

    with pytest.raises(RuntimeError) as e:
        param.set(1)
    assert str(e.value) == "('Trying to set Group value but no group defined', 'setting b to 1')"


def test_raises_runtime_error_on_update_if_get_cmd_is_none():
    dummy = Dummy("dummy", get_cmd=None)
    msg = ("Cannot update values in the group with "
           "parameters - dummy_a, dummy_b since it "
           "has no `get_cmd` defined.")
    with pytest.raises(RuntimeError, match=msg):
        dummy.group.update()

def test_raises_runtime_error_if_set_parameters_called_with_empty_dict():
    dummy = Dummy("dummy")
    parameters_dict = dict()
    msg = ("Provide at least one group parameter and its value to be set.")

    with pytest.raises(RuntimeError, match=msg):
        dummy.group.set_parameters(parameters_dict)

def test_set_parameters_called_for_one_parameter():
    dummy = Dummy("dummy")
    parameters_dict = {"a": 7}

    dummy.group.set_parameters(parameters_dict)
    assert dummy.a() == 7
    assert dummy.b() == 0

def test_set_parameters_called_for_more_than_one_parameters():
    dummy = Dummy("dummy")
    parameters_dict = {"a": 10, "b": 57}

    dummy.group.set_parameters(parameters_dict)
    assert dummy.a() == 10
    assert dummy.b() == 57

def test_set_parameters_when_parameter_value_not_equal_to_raw_value():
    dummy = Dummy("dummy", scale_a=10)
    parameters_dict = {"a": 7}

    dummy.group.set_parameters(parameters_dict)
    assert dummy.a.cache.get(get_if_invalid=False) == 7
    assert dummy.a.cache.raw_value == 70
    assert dummy.a() == 7

def test_initial_values():
    initial_a = 42
    initial_b = 43
    dummy = Dummy("dummy", initial_a=initial_a, initial_b=initial_b)

    assert dummy.a() == initial_a
    assert dummy.b() == initial_b


def test_raise_on_not_all_initial_values():
    expected_err_msg = (r'Either none or all of the parameters in a group '
                        r'should have an initial value. Found initial values '
                        r'for \[.*\] but not for \[.*\].')
    with pytest.raises(ValueError, match=expected_err_msg):
        dummy = Dummy("dummy", initial_a=42)


def test_update_group_parameter_reflected_in_cache_of_all_params():
    dummy = Dummy("dummy")
    group = dummy.a.group

    assert dummy.a.cache.timestamp is None
    assert dummy.b.cache.timestamp is None
    assert dummy.a.cache.get(get_if_invalid=False) is None
    assert dummy.b.cache.get(get_if_invalid=False) is None

    before = datetime.now()
    group.update()
    after = datetime.now()

    assert before <= dummy.a.cache.timestamp
    assert after >= dummy.a.cache.timestamp

    assert before <= dummy.b.cache.timestamp
    assert after >= dummy.b.cache.timestamp

    assert dummy.a.cache.get(get_if_invalid=False) == 0
    assert dummy.b.cache.get(get_if_invalid=False) == 0


def test_get_group_param_updates_cache_of_other_param():
    dummy = Dummy("dummy")

    assert dummy.a.cache.timestamp is None
    assert dummy.b.cache.timestamp is None
    assert dummy.a.cache.get(get_if_invalid=False) is None
    assert dummy.b.cache.get(get_if_invalid=False) is None

    before = datetime.now()
    assert dummy.a.get() == 0
    after = datetime.now()

    assert before <= dummy.a.cache.timestamp
    assert after >= dummy.a.cache.timestamp

    assert before <= dummy.b.cache.timestamp
    assert after >= dummy.b.cache.timestamp

    assert dummy.a.cache.get(get_if_invalid=False) == 0
    assert dummy.b.cache.get(get_if_invalid=False) == 0


def test_set_group_param_updates_cache_of_other_param():
    dummy = Dummy("dummy")

    assert dummy.a.cache.timestamp is None
    assert dummy.b.cache.timestamp is None
    assert dummy.a.cache.get(get_if_invalid=False) is None
    assert dummy.b.cache.get(get_if_invalid=False) is None

    before = datetime.now()
    dummy.a.set(10)
    after = datetime.now()

    assert before <= dummy.a.cache.timestamp
    assert after >= dummy.a.cache.timestamp

    assert before <= dummy.b.cache.timestamp
    assert after >= dummy.b.cache.timestamp

    assert dummy.a.cache.get(get_if_invalid=False) == 10
    assert dummy.b.cache.get(get_if_invalid=False) == 0


def test_group_param_scale_is_handled():
    dummy = Dummy("dummy", scale_a=10, initial_a=1, initial_b=5)

    assert dummy.a.cache.get(get_if_invalid=False) == 1
    assert dummy.a.cache.raw_value == 10
    assert dummy.a.get() == 1

    dummy.a.set(10)

    assert dummy.a.cache.get(get_if_invalid=False) == 10
    assert dummy.a.cache.raw_value == 100



from hypothesis import given
import hypothesis.strategies as st
from qcodes.instrument_drivers.Lakeshore import Model_325

STATUS = Model_325.Status


@given(
    st.lists(st.sampled_from(list(STATUS)),
             min_size=1, max_size=5, unique=True).map(sorted)
)
def test_decode_sensor_status(list_of_codes):
    """
    The sensor status is one of the status codes, or a sum thereof. Multiple
    status are possible as they are not necessarily mutually exclusive.
    The static method 'decode_sensor_status' in the Model_325_Sensor class can
    decode the status into status code(s).
    """
    codes = [code.name.replace('_', ' ') for code in list_of_codes[::-1]]
    codes_message = ', '.join(codes)
    sum_of_codes = int(sum(list_of_codes))
    status_messages = Model_325.Model_325_Sensor.decode_sensor_status(
        sum_of_codes
    )
    assert codes_message == status_messages


def test_decode_sensor_status_0():
    status_codes = Model_325.Model_325_Sensor.decode_sensor_status(0)
    assert status_codes == 'OK'
